# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Functions to export the rig part of colmap in json format.
"""

import json
import logging
import os.path as path
from typing import Dict, List

# kapture
import kapture

logger = logging.getLogger('colmap')


def export_colmap_rig_json(kapture_rigs: kapture.Rigs,
                           records_camera: kapture.RecordsCamera,
                           colmap_camera_ids: Dict[str, int]
                           ) -> List[Dict[str, list]]:
    """
    From colmap source code comments (colmap.cc:1401) :

    // Read the configuration of the camera rigs from a JSON file. The input images
    // of a camera rig must be named consistently to assign them to the appropriate
    // camera rig and the respective kapture image.
    //
    // An example configuration of a single camera rig:
    // [
    //   {
    //     "ref_camera_id": 1,
    //     "cameras":
    //     [
    //       {
    //           "camera_id": 1,
    //           "image_prefix": "left1_image"
    //       }, ...
    //     ]
    //   }
    // ]


    // This file specifies the configuration for a single camera rig and that you
    // could potentially define multiple camera rigs.


    Images with the same prefix ("left1_image/") are assigned to the same camera in the rig.
    Images with the same suffix ("_frame000.png" and "/frame000.png") are
    assigned to the same camera record, i.e., they are assumed to be captured at the same time.

    This code, try to retrieve prefix of images.

    :param kapture_rigs: kapture rigs to export
    :param records_camera: used to guess camera prefix: assume a directory per camera
    :param colmap_camera_ids: dict mapping sensor_id -> colmap_camera_ID
    :return: The structure as expected in the colmap json file.
    """

    colmap_rigs = []
    for rig_kapture in kapture_rigs.values():
        # rig[sensor_device_id] = <Pose>
        rig_colmap = {'cameras': []}
        for camera_id in rig_kapture:
            if camera_id not in colmap_camera_ids:
                # e.g. lidar0 and other sensors without record on disk
                logger.warning(f'Camera {camera_id} not in COLMAP cameras')
                continue
            colmap_camera_id = colmap_camera_ids[camera_id]
            # recover the image prefix :
            images_dir_paths = set(path.dirname(record_filepath)
                                   for _, sensor_id, record_filepath in kapture.flatten(records_camera)
                                   if sensor_id == camera_id)
            # check there is one, and only one image prefix (required by colmap).
            if not len(images_dir_paths) == 1:
                raise ValueError(f'unable to find an image for camera {camera_id}')

            the_images_dirpath = next(iter(images_dir_paths))
            if not the_images_dirpath:
                raise ValueError(f'unable to find prefix for images {the_images_dirpath}.')
            rig_colmap['cameras'].append({
                'camera_id': colmap_camera_id,
                'image_prefix': the_images_dirpath
            })
        rig_colmap['ref_camera_id'] = rig_colmap['cameras'][0]['camera_id']
        colmap_rigs.append(rig_colmap)

    return colmap_rigs


def export_colmap_rig(colmap_rig_filepath: str,
                      rigs: kapture.Rigs,
                      records_camera: kapture.RecordsCamera,
                      colmap_camera_ids: dict) -> None:
    """
    Exports kapture rigs to colmap rigs file.

    :param colmap_rig_filepath: file path where to write rig.
    :param rigs: kapture rigs.
    :param records_camera: used to guess camera prefix: assume a directory per camera
    :param colmap_camera_ids: dict mapping colmap_camera_ID -> sensor_id
    """
    colmap_rigs = export_colmap_rig_json(rigs, records_camera, colmap_camera_ids)
    with open(colmap_rig_filepath, 'w') as outfile:
        json.dump(colmap_rigs, outfile, indent=4, sort_keys=True)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

from kapture.utils.Collections import AutoEnum
from enum import auto
from typing import Union, Optional, List, Dict


class Sensor:
    """
    brief:  name, sensor_type, [sensor_params]+
            sensor_params is a list of string, comma separated.
    """

    def __init__(self, sensor_type: str, sensor_params: Optional[list] = None, name: Optional[str] = None):
        assert isinstance(sensor_type, str)
        assert sensor_params is None or isinstance(sensor_params, list)
        assert name is None or isinstance(name, str)
        self.name = name
        self._sensor_type = sensor_type
        self.sensor_params = sensor_params or []

    @property
    def sensor_type(self):
        """
        :return: sensor type as string
        """
        return self._sensor_type

    def __repr__(self) -> str:
        representation = ''
        representation += f'name: {self.name or "--":5} '
        representation += f'type: {self.sensor_type:6} '
        representation += '[{}]'.format(', '.join(f'{i:3}' for i in self.sensor_params))
        return representation


class Sensors(Dict[str, Sensor]):
    """
    brief: sensors[sensor_id] = <sensor>
    """

    def __setitem__(self, sensor_id: str, sensor: Sensor):
        # enforce type checking
        if not isinstance(sensor_id, str):
            raise TypeError('invalid type for sensor_id')
        if not isinstance(sensor, Sensor):
            raise TypeError('invalid type of sensor')
        super(Sensors, self).__setitem__(sensor_id, sensor)

    def __repr__(self) -> str:
        # [sensor_id] = name, type, [params]+
        representation = '\n'.join(
            f'[{sensor_id:5}] = {sensor}' for sensor_id, sensor in self.items()
        )
        return representation


class CameraType(AutoEnum):
    """
    Enumeration that contains all the supported camera types
    """
    SIMPLE_PINHOLE = auto()
    PINHOLE = auto()
    SIMPLE_RADIAL = auto()
    RADIAL = auto()
    OPENCV = auto()
    OPENCV_FISHEYE = auto()
    FULL_OPENCV = auto()
    FOV = auto()
    SIMPLE_RADIAL_FISHEYE = auto()
    RADIAL_FISHEYE = auto()
    THIN_PRISM_FISHEYE = auto()
    UNKNOWN_CAMERA = auto()


CAMERA_TYPE_PARAMS_COUNT = {
    # https://en.wikipedia.org/wiki/Pinhole_camera_model
    # w, h, f, cx, cy
    CameraType.SIMPLE_PINHOLE: 5,

    # https://en.wikipedia.org/wiki/Pinhole_camera_model
    # w, h, fx, fy, cx, cy
    CameraType.PINHOLE: 6,

    # Simplified versions of the OPENCV model only modeling radial distortion.
    # w, h, f, cx, cy, k
    # equivalent to CameraType.RADIAL with k1=k, k2=0
    # see: https://github.com/colmap/colmap/blob/master/src/base/camera_models.h#L715
    CameraType.SIMPLE_RADIAL: 6,

    # Simplified versions of the OPENCV model only modeling radial distortion
    # w, h, f, cx, cy, k1, k2
    # r^2 = u^2 + v^2;
    # radial = k1 * r^2 + k2 * r^2 * r^2
    # u_distorted = u + (u * radial)
    # where u,v are normalized camera coordinates
    # see: https://github.com/colmap/colmap/blob/master/src/base/camera_models.h#L784
    CameraType.RADIAL: 7,

    # http://docs.opencv.org/modules/calib3d/doc/camera_calibration_and_3d_reconstruction.html
    # w, h, fx, fy, cx, cy, k1, k2, p1, p2
    CameraType.OPENCV: 10,

    # http://docs.opencv.org/modules/calib3d/doc/camera_calibration_and_3d_reconstruction.html
    # w, h, fx, fy, cx, cy, k1, k2, k3, k4
    CameraType.OPENCV_FISHEYE: 10,

    # http://docs.opencv.org/modules/calib3d/doc/camera_calibration_and_3d_reconstruction.html
    # w, h, fx, fy, cx, cy, k1, k2, p1, p2, k3, k4, k5, k6
    CameraType.FULL_OPENCV: 14,

    # Frederic Devernay, Olivier Faugeras. Straight lines have to be straight:
    # Automatic calibration and removal of distortion from scenes of structured
    # environments. Machine vision and applications, 2001.
    # w, h, fx, fy, cx, cy, omega
    CameraType.FOV: 7,

    # equivalent to the OpenCVFisheyeCameraModel but has only one
    # radial distortion coefficient.
    # w, h, f, cx, cy, k
    CameraType.SIMPLE_RADIAL_FISHEYE: 6,

    # equivalent to the OpenCVFisheyeCameraModel but has only two
    # radial distortion coefficients.
    # w, h, f, cx, cy, k1, k2
    CameraType.RADIAL_FISHEYE: 7,

    # Camera Calibration with Distortion Models and Accuracy Evaluation,
    # J Weng et al., TPAMI, 1992.
    # w, h, fx, fy, cx, cy, k1, k2, p1, p2, k3, k4, sx1, sx2
    CameraType.THIN_PRISM_FISHEYE: 14,

    # Non calibrated camera
    # w, h
    CameraType.UNKNOWN_CAMERA: 2
}

CAMERA_TYPE_PARAMS_COUNT_FROM_NAME = {
    field.value: n
    for field, n in CAMERA_TYPE_PARAMS_COUNT.items()
}

SENSOR_TYPE_CAMERA = 'camera'
SENSOR_TYPE_DEPTH_CAM = 'depth'
ALL_CAMERA_SENSOR_TYPES = [SENSOR_TYPE_CAMERA, SENSOR_TYPE_DEPTH_CAM]


class Camera(Sensor):
    """
    A camera definition
    """

    def __init__(self,
                 camera_type: Union[CameraType, str],
                 camera_params: list = None,
                 name: Optional[str] = None,
                 sensor_type: str = SENSOR_TYPE_CAMERA):
        # type checking
        assert name is None or isinstance(name, str)
        if isinstance(camera_type, str):
            camera_type = CameraType[camera_type]
        assert isinstance(camera_type, CameraType)
        # check params are consistent with model
        assert isinstance(camera_params, list)
        assert len(camera_params) == CAMERA_TYPE_PARAMS_COUNT[camera_type]
        assert sensor_type in ALL_CAMERA_SENSOR_TYPES

        # make sure it crashes if camera_params cannot be cast to float, store as string in sensor_params
        camera_params = [float(v) for v in camera_params]
        camera_params = [str(int(v)) if v.is_integer() else str(v) for v in camera_params]
        sensor_params = [camera_type.name] + camera_params
        super(Camera, self).__init__(sensor_type=sensor_type, sensor_params=sensor_params, name=name)

    @property
    def camera_type(self) -> CameraType:
        """
        :return: the camera type
        """
        if self.sensor_params is None:
            raise ValueError('sensor_params should not be None for a camera')
        return CameraType[self.sensor_params[0]]

    @property
    def camera_params(self) -> List[float]:
        """
        :return: camera parameters
        """
        if self.sensor_params is None:
            raise ValueError('sensor_params should not be None for a camera')
        num_params = CAMERA_TYPE_PARAMS_COUNT[self.camera_type]
        return [float(c) for c in self.sensor_params[1:(1+num_params)]]


def create_sensor(sensor_type: str, sensor_params: Optional[list] = None, name: Optional[str] = None):
    """
    Creates a instance of a sensor

    :param sensor_type: type of sensor ('camera', ...)
    :param sensor_params: sensor specific parameters
    :param name: sensor name
    :return: created instance
    """
    if sensor_type in ALL_CAMERA_SENSOR_TYPES:
        assert sensor_params is not None
        return Camera(camera_type=sensor_params[0], camera_params=sensor_params[1:], name=name, sensor_type=sensor_type)
    else:
        return Sensor(sensor_type, sensor_params, name)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This files contains IO operations on Record related data.
"""

import os
import os.path as path
from typing import Dict, Optional, Union, Iterable, Tuple
import numpy as np
import kapture
from kapture.utils.paths import path_secure
from kapture.utils.logging import getLogger
from kapture.io.binary import TransferAction, transfer_files_from_dir, array_from_file, array_to_file

logger = getLogger()

# Records files related functions ######################################################################################
RECORD_DATA_DIRNAME = path_secure(path.join('sensors', 'records_data'))


########################################################################################################################
def get_record_fullpath(
        kapture_dirpath: str = '',
        record_filename: Optional[str] = None) -> str:
    """
    Returns full path to subdirectory containing the binary files of the record type.
            Optionally, can give a the file name.
    :param kapture_dirpath: input path to kapture directory.
    :param record_filename: optional input record filename (eg. image filename).
    :return: the record full path
    """
    feature_filename = record_filename or ''
    return path_secure(path.join(kapture_dirpath, RECORD_DATA_DIRNAME, feature_filename))


def records_to_filepaths(
        records: Union[kapture.RecordsCamera,
                       kapture.RecordsDepth,
                       kapture.RecordsWifi,
                       kapture.RecordsLidar,
                       kapture.RecordsGnss],
        kapture_dirpath: str
) -> Dict[str, str]:
    """
    Computes filepaths for records.

    :param records: records
    :param kapture_dirpath: top kapture directory path
    :return: records name to records file path dictionary
    """
    return {filename: path_secure(path.join(kapture_dirpath, RECORD_DATA_DIRNAME, filename))
            for _, _, filename in kapture.flatten(records)}


def guess_filepaths_from_filenames(
        dirpath: str,
        filenames: Iterable[str]
):
    """ returns a generator that prepend the directory path to the given filenames."""
    return (path_secure(path.join(dirpath, record_filename))
            for record_filename in filenames)


def import_record_data_from_dir_link_dir(
        source_record_dirpath: str,
        kapture_dirpath: str,
        do_relative_link: bool = False
) -> None:
    """
    Imports record_data from a given directory:
    This do not actually copy any files, but make a symbolic link on the root path instead.

    :param source_record_dirpath: input path to directory where to import data from.
    :param kapture_dirpath: input root path to kapture root directory, where to import data to.
    :param do_relative_link: if True, do relative link else absolute link.
    """
    assert path.isdir(source_record_dirpath)
    kapture_record_dirpath = get_record_fullpath(kapture_dirpath)
    if do_relative_link:
        source_record_dirpath = path.relpath(source_record_dirpath, kapture_record_dirpath)
    try:  # on windows, symlink requires some privileges, and may crash if not
        os.symlink(source_record_dirpath, kapture_record_dirpath)
    except OSError as e:
        logger.critical('unable to create symlink on image directory, due to privilege restrictions.')
        raise e


def import_record_data_from_dir_auto(
        source_record_dirpath: str,
        destination_kapture_dirpath: str,
        filename_list: Iterable[str],
        copy_strategy: TransferAction = TransferAction.copy
) -> None:
    """
    Imports record_data from a given directory.
    Automatically replicate the folder hierarchy from source to kapture record directory.
    If you want to change the file organisation use import_record_data_from_dir_explicit instead.
    The actual import can be done using actual file copy (copy/move), or symlinks (absolute/relative).

    :param source_record_dirpath: input path to directory where to import data from.
    :param destination_kapture_dirpath: input root path to kapture root directory, where to import data to.
    :param filename_list: input list of filenames to import (filenames are relative path to source_record_dirpath).
                    This image list can be obtained from file walking (populate_files_in_dirpath)
                    or using an already populated kapture.RecordsData. Prefer RecordsData since it will only copy
                    required files, eg:
                    filename_list = [f for _, _, f in kapture.flatten(kapture_data.records_camera)]
    :param copy_strategy:
    """
    if copy_strategy == TransferAction.root_link:
        import_record_data_from_dir_link_dir(source_record_dirpath, destination_kapture_dirpath)
    else:
        source_filepath_list = (path.join(source_record_dirpath, record_filename)
                                for record_filename in filename_list)
        kapture_filepath_list = (get_record_fullpath(destination_kapture_dirpath, record_filename)
                                 for record_filename in filename_list)
        transfer_files_from_dir(
            source_filepath_list,
            kapture_filepath_list,
            copy_strategy
        )


# images ###############################################################################################################
def get_image_fullpath(kapture_dir_path: str, image_filename: Optional[str] = None) -> str:
    """
    Get the full path of the image file in the kapture.
     If image_filename is missing, this gives the top directory (records_data).

    :param kapture_dir_path: the kapture top directory
    :param image_filename: optional image file name
    :return: Image file full path
    """
    return get_record_fullpath(kapture_dir_path, image_filename)


def images_to_filepaths(images: kapture.RecordsCamera, kapture_dirpath: str) -> Dict[str, str]:
    """
    Computes filepaths for image records.

    :param images: images records
    :param kapture_dirpath: top kapture directory path
    :return: images name to images file path dictionary
    """
    return records_to_filepaths(images, kapture_dirpath)


# depth maps file paths ################################################################################################
def get_depth_map_fullpath(kapture_dir_path: str, depth_map_filename: Optional[str] = None) -> str:
    """
    Get the full path of the depth map file in the kapture.
     If depth_map_filename is missing, this gives the top directory (records_data).

    :param kapture_dir_path: the kapture top directory
    :param depth_map_filename: optional image file name
    :return: Depth map file full path
    """
    return get_record_fullpath(kapture_dir_path, depth_map_filename)


def depth_maps_to_filepaths(depth_records: kapture.RecordsDepth, kapture_dirpath: str) -> Dict[str, str]:
    """
    Computes filepaths for depth maps records.

    :param images: images records
    :param kapture_dirpath: top kapture directory path
    :return: images name to images file path dictionary
    """
    return records_to_filepaths(depth_records, kapture_dirpath)


# depth maps IO ########################################################################################################
def records_depth_from_file(filepath: str, size: Tuple[int, int]) -> np.array:
    """
    Load the depth map from binary file to a numpy array.

    :param filepath: path to the file
    :param size: [width, height]
    :return: the depth map as a numpy array
    """
    assert isinstance(size, tuple) and len(size) == 2
    dtype = kapture.RecordsDepth.dtype
    dsize = int(size[0] * size[1])
    bitmap = array_from_file(filepath, dtype, dsize)
    bitmap = bitmap.reshape((size[1], size[0]))
    return bitmap


def records_depth_to_file(filepath: str, depth_map: np.array) -> None:
    """
    Writes the depth map from numpy array to binary file.

    :param filepath: file path
    :param depth_map: depth map as a numpy array
    """
    array_to_file(filepath, depth_map)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
OpenSfM to kapture import functions.
"""

import logging
import os
import os.path as path
import numpy as np
import quaternion
import gzip
import pickle
import json
from tqdm import tqdm
from typing import Any, Dict, Optional, Tuple
# kapture
import kapture
from kapture.io.csv import kapture_to_dir
import kapture.io.features
from kapture.io.records import TransferAction, import_record_data_from_dir_auto
from kapture.io.structure import delete_existing_kapture_files

logger = logging.getLogger('opensfm')

"""
opensfm_project/
├── config.yaml
├── images/
├── masks/
├── gcp_list.txt
├── exif/
├── camera_models.json
├── features/
├── matches/
├── tracks.csv
├── reconstruction.json
├── reconstruction.meshed.json
└── undistorted/
    ├── images/
    ├── masks/
    ├── tracks.csv
    ├── reconstruction.json
    └── depthmaps/
        └── merged.ply
"""

"""
reconstruction.json: [RECONSTRUCTION, ...]

RECONSTRUCTION: {
    "cameras": {
        CAMERA_ID: CAMERA,
        ...
    },
    "shots": {
        SHOT_ID: SHOT,
        ...
    },
    "points": {
        POINT_ID: POINT,
        ...
    }
}

CAMERA: {
    "projection_type": "perspective",  # Can be perspective, brown, fisheye or equirectangular
    "width": NUMBER,                   # Image width in pixels
    "height": NUMBER,                  # Image height in pixels

    # Depending on the projection type more parameters are stored.
    # These are the parameters of the perspective camera.
    "focal": NUMBER,                   # Estimated focal length
    "k1": NUMBER,                      # Estimated distortion coefficient
    "k2": NUMBER,                      # Estimated distortion coefficient
}

SHOT: {
    "camera": CAMERA_ID,
    "rotation": [X, Y, Z],      # Estimated rotation as an angle-axis vector
    "translation": [X, Y, Z],   # Estimated translation
    "gps_position": [X, Y, Z],  # GPS coordinates in the reconstruction reference frame
    "gps_dop": METERS,          # GPS accuracy in meters
    "orientation": NUMBER,      # EXIF orientation tag (can be 1, 3, 6 or 8)
    "capture_time": SECONDS     # Capture time as a UNIX timestamp
}

POINT: {
    "coordinates": [X, Y, Z],      # Estimated position of the point
    "color": [R, G, B],            # Color of the point
}
"""


def import_camera(
        opensfm_camera: Dict[str, Any],
        name: Optional[str] = None
) -> kapture.Camera:
    """
    Converts OpenSfM camera to kapture.

    :param opensfm_camera: openSfM camera definition in a dictionary
    :param name: camera name
    :return: kapture camera definition
    """

    # opensfm_camera['projection_type'] can be perspective, brown, fisheye or equirectangular

    if 'perspective' == opensfm_camera['projection_type']:
        # convert to CameraType.RADIAL [w, h, f, cx, cy, k1, k2]
        # missing principal point, just fake it at image center
        largest_side_in_pixel = float(max(opensfm_camera['width'], opensfm_camera['height']))
        camera_params = [
            # w, h:
            opensfm_camera['width'], opensfm_camera['height'],
            # f: The focal length provided by the EXIF metadata divided by the sensor width
            opensfm_camera['focal'] * largest_side_in_pixel,
            # cx, cy: no principal point, guess one at image center
            opensfm_camera['width'] / 2, opensfm_camera['height'] / 2,
            # k1, k2
            opensfm_camera.get('k1', 0.0), opensfm_camera.get('k2', 0.0),
        ]
        return kapture.Camera(
            camera_type=kapture.CameraType.RADIAL,
            camera_params=camera_params,
            name=name
        )

    else:
        raise ValueError(f'unable to convert camera of type {opensfm_camera["projection_type"]}')


def _import_gnss(opensfm_root_dir, kapture_sensors, image_sensors, image_timestamps, disable_tqdm) \
        -> Optional[kapture.RecordsGnss]:
    """
    Imports the GNSS info from the images exif.

    """
    # gps from pre-extracted exif, in exif/image_name.jpg.exif
    kapture_gnss = None
    opensfm_exif_dir_path = path.join(opensfm_root_dir, 'exif')
    opensfm_exif_suffix = '.exif'
    if path.isdir(opensfm_exif_dir_path):
        logger.info('importing GNSS from exif ...')
        camera_ids = set(image_sensors.values())
        # add a gps sensor for each camera
        map_cam_to_gnss_sensor = {cam_id: 'GPS_' + cam_id for cam_id in camera_ids}
        for gnss_id in map_cam_to_gnss_sensor.values():
            kapture_sensors[gnss_id] = kapture.Sensor(sensor_type='gnss', sensor_params=['EPSG:4326'])
        # build epsg_code for all cameras
        kapture_gnss = kapture.RecordsGnss()
        opensfm_exif_filepath_list = (path.join(dir_path, filename)
                                      for dir_path, _, filename_list in os.walk(opensfm_exif_dir_path)
                                      for filename in filename_list
                                      if filename.endswith(opensfm_exif_suffix))
        for opensfm_exif_filepath in tqdm(opensfm_exif_filepath_list, disable=disable_tqdm):
            image_filename = path.relpath(opensfm_exif_filepath, opensfm_exif_dir_path)[:-len(opensfm_exif_suffix)]
            image_timestamp = image_timestamps[image_filename]
            image_sensor_id = image_sensors[image_filename]
            gnss_timestamp = image_timestamp
            gnss_sensor_id = map_cam_to_gnss_sensor[image_sensor_id]
            with open(opensfm_exif_filepath, 'rt') as f:
                js_root = json.load(f)
                if 'gps' not in js_root:
                    logger.warning(f'NO GPS data in "{opensfm_exif_filepath}"')
                    continue

                gps_coords = {
                    'x': js_root['gps']['longitude'],
                    'y': js_root['gps']['latitude'],
                    'z': js_root['gps'].get('altitude', 0.0),
                    'dop': js_root['gps'].get('dop', 0),
                    'utc': 0,
                }
                logger.debug(f'found GPS data for ({gnss_timestamp}, {gnss_sensor_id}) in "{opensfm_exif_filepath}"')
                kapture_gnss[gnss_timestamp, gnss_sensor_id] = kapture.RecordGnss(**gps_coords)
    return kapture_gnss


def _import_features_and_matches(opensfm_root_dir, kapture_root_dir, disable_tqdm,
                                 keypoints_type: str = 'HessianAffine',
                                 descriptors_type: str = 'HOG')\
        -> Tuple[Optional[kapture.Descriptors],
                 Optional[kapture.Keypoints],
                 Optional[kapture.Matches]]:
    # import features (keypoints + descriptors)
    kapture_keypoints = None  # kapture.Keypoints(type_name='opensfm', dsize=4, dtype=np.float64)
    kapture_descriptors = None  # kapture.Descriptors(type_name='opensfm', dsize=128, dtype=np.uint8)
    opensfm_features_dir_path = path.join(opensfm_root_dir, 'features')
    opensfm_features_suffix = '.features.npz'
    if path.isdir(opensfm_features_dir_path):
        logger.info('importing keypoints and descriptors ...')
        opensfm_features_file_list = (path.join(dp, fn)
                                      for dp, _, fs in os.walk(opensfm_features_dir_path) for fn in fs)
        opensfm_features_file_list = (filepath
                                      for filepath in opensfm_features_file_list
                                      if filepath.endswith(opensfm_features_suffix))
        for opensfm_feature_filename in tqdm(opensfm_features_file_list, disable=disable_tqdm):
            image_filename = path.relpath(opensfm_feature_filename, opensfm_features_dir_path)[
                :-len(opensfm_features_suffix)]
            opensfm_image_features = np.load(opensfm_feature_filename)
            opensfm_image_keypoints = opensfm_image_features['points']
            opensfm_image_descriptors = opensfm_image_features['descriptors']
            logger.debug(f'parsing keypoints and descriptors in {opensfm_feature_filename}')
            if kapture_keypoints is None:
                # print(type(opensfm_image_keypoints.dtype))
                # HAHOG = Hessian Affine feature point detector + HOG descriptor
                kapture_keypoints = kapture.Keypoints(
                    type_name=keypoints_type,
                    dsize=opensfm_image_keypoints.shape[1],
                    dtype=opensfm_image_keypoints.dtype)
            if kapture_descriptors is None:
                kapture_descriptors = kapture.Descriptors(
                    type_name=descriptors_type,
                    dsize=opensfm_image_descriptors.shape[1],
                    dtype=opensfm_image_descriptors.dtype,
                    keypoints_type=keypoints_type,
                    metric_type='L2')

            # convert keypoints file
            keypoint_file_path = kapture.io.features.get_features_fullpath(
                data_type=kapture.Keypoints,
                feature_type=keypoints_type,
                kapture_dirpath=kapture_root_dir,
                image_filename=image_filename)
            kapture.io.features.image_keypoints_to_file(
                filepath=keypoint_file_path, image_keypoints=opensfm_image_keypoints)
            # register the file
            kapture_keypoints.add(image_filename)

            # convert descriptors file
            descriptor_file_path = kapture.io.features.get_features_fullpath(
                data_type=kapture.Descriptors,
                feature_type=descriptors_type,
                kapture_dirpath=kapture_root_dir,
                image_filename=image_filename)
            kapture.io.features.image_descriptors_to_file(
                filepath=descriptor_file_path, image_descriptors=opensfm_image_descriptors)
            # register the file
            kapture_descriptors.add(image_filename)
    # import matches
    kapture_matches = kapture.Matches()
    opensfm_matches_suffix = '_matches.pkl.gz'
    opensfm_matches_dir_path = path.join(opensfm_root_dir, 'matches')
    if path.isdir(opensfm_matches_dir_path):
        logger.info('importing matches ...')
        opensfm_matches_file_list = (path.join(dp, fn)
                                     for dp, _, fs in os.walk(opensfm_matches_dir_path) for fn in fs)
        opensfm_matches_file_list = (filepath
                                     for filepath in opensfm_matches_file_list
                                     if filepath.endswith(opensfm_matches_suffix))

        for opensfm_matches_filename in tqdm(opensfm_matches_file_list, disable=disable_tqdm):
            image_filename_1 = path.relpath(opensfm_matches_filename, opensfm_matches_dir_path)[
                :-len(opensfm_matches_suffix)]
            logger.debug(f'parsing matches in {image_filename_1}')
            with gzip.open(opensfm_matches_filename, 'rb') as f:
                opensfm_matches = pickle.load(f)
                for image_filename_2, opensfm_image_matches in opensfm_matches.items():
                    image_pair = (image_filename_1, image_filename_2)
                    # register the pair to kapture
                    kapture_matches.add(*image_pair)
                    # convert the bin file to kapture
                    kapture_matches_filepath = kapture.io.features.get_matches_fullpath(
                        image_filename_pair=image_pair,
                        keypoints_type=keypoints_type,
                        kapture_dirpath=kapture_root_dir)
                    kapture_image_matches = np.hstack([
                        opensfm_image_matches.astype(np.float64),
                        # no matches scoring = assume all to one
                        np.ones(shape=(opensfm_image_matches.shape[0], 1), dtype=np.float64)])
                    kapture.io.features.image_matches_to_file(kapture_matches_filepath, kapture_image_matches)
    return kapture_descriptors, kapture_keypoints, kapture_matches


def import_opensfm(
        opensfm_root_dir: str,
        kapture_root_dir: str,
        force_overwrite_existing: bool = False,
        images_import_method: TransferAction = TransferAction.copy,
        keypoints_type: str = 'HessianAffine',
        descriptors_type: str = 'HOG'
) -> None:
    """
    Convert an openSfM structure to a kapture on disk. Also copy, move or link the images files if necessary.

    :param opensfm_root_dir: the openSfM top directory
    :param kapture_root_dir: top directory of kapture created
    :param force_overwrite_existing: if true, will remove existing kapture data without prompting the user
    :param images_import_method: action to apply on images: link, copy, move or do nothing.
    :return: the constructed kapture object
    """
    disable_tqdm = logger.getEffectiveLevel() != logging.INFO
    # load reconstruction
    opensfm_reconstruction_filepath = path.join(opensfm_root_dir, 'reconstruction.json')
    with open(opensfm_reconstruction_filepath, 'rt') as f:
        opensfm_reconstruction = json.load(f)
    # remove the single list @ root
    opensfm_reconstruction = opensfm_reconstruction[0]

    # prepare space for output
    os.makedirs(kapture_root_dir, exist_ok=True)
    delete_existing_kapture_files(kapture_root_dir, force_erase=force_overwrite_existing)

    # import cameras
    kapture_sensors = kapture.Sensors()
    assert 'cameras' in opensfm_reconstruction
    # import cameras
    for osfm_camera_id, osfm_camera in opensfm_reconstruction['cameras'].items():
        camera = import_camera(osfm_camera, name=osfm_camera_id)
        kapture_sensors[osfm_camera_id] = camera

    # import shots
    logger.info('importing images and trajectories ...')
    kapture_images = kapture.RecordsCamera()
    kapture_trajectories = kapture.Trajectories()
    opensfm_image_dir_path = path.join(opensfm_root_dir, 'images')
    assert 'shots' in opensfm_reconstruction
    image_timestamps, image_sensors = {}, {}  # used later to retrieve the timestamp of an image.
    for timestamp, (image_filename, shot) in enumerate(opensfm_reconstruction['shots'].items()):
        sensor_id = shot['camera']
        image_timestamps[image_filename] = timestamp
        image_sensors[image_filename] = sensor_id
        # in OpenSfm, (sensor, timestamp) is not unique.
        rotation_vector = shot['rotation']
        q = quaternion.from_rotation_vector(rotation_vector)
        translation = shot['translation']
        # capture_time = shot['capture_time'] # may be invalid
        # gps_position = shot['gps_position']
        kapture_images[timestamp, sensor_id] = image_filename
        kapture_trajectories[timestamp, sensor_id] = kapture.PoseTransform(r=q, t=translation)

    # copy image files
    filename_list = [f for _, _, f in kapture.flatten(kapture_images)]
    import_record_data_from_dir_auto(
        source_record_dirpath=opensfm_image_dir_path,
        destination_kapture_dirpath=kapture_root_dir,
        filename_list=filename_list,
        copy_strategy=images_import_method)

    # Imports Gnss
    kapture_gnss = _import_gnss(opensfm_root_dir, kapture_sensors, image_sensors, image_timestamps, disable_tqdm)
    # Imports descriptors, keypoints and matches
    kapture_descriptors, kapture_keypoints, kapture_matches = _import_features_and_matches(opensfm_root_dir,
                                                                                           kapture_root_dir,
                                                                                           disable_tqdm,
                                                                                           keypoints_type,
                                                                                           descriptors_type)

    # import 3-D points
    if 'points' in opensfm_reconstruction:
        logger.info('importing points 3-D')
        opensfm_points = opensfm_reconstruction['points']
        points_data = []
        for point_id in sorted(opensfm_points):
            point_data = opensfm_points[point_id]
            point_data = point_data['coordinates'] + point_data['color']
            points_data.append(point_data)
        kapture_points = kapture.Points3d(points_data)
    else:
        kapture_points = None

    # saving kapture csv files
    logger.info('saving kapture files')
    kapture_data = kapture.Kapture(
        sensors=kapture_sensors,
        records_camera=kapture_images,
        records_gnss=kapture_gnss,
        trajectories=kapture_trajectories,
        keypoints={keypoints_type: kapture_keypoints} if kapture_keypoints is not None else None,
        descriptors={descriptors_type: kapture_descriptors} if kapture_descriptors is not None else None,
        matches={keypoints_type: kapture_matches} if kapture_matches is not None else None,
        points3d=kapture_points
    )
    kapture_to_dir(kapture_root_dir, kapture_data)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
The trajectories points. We store for every timestamp and device a pose.
So you can have two different devices for the same timestamp with a different pose.
The timestamps are integers, and device identifiers strings.
The timestamps are often epoch timestamps (in micro or milliseconds), but can be any set of integers.
Note that the timestamps are usually ordered ascending (as the time goes), but this is not guaranteed,
and might change if you manipulate them, save and reload them from disk.
If you need them ordered, do it yourself.
"""

import logging
import quaternion
from tqdm import tqdm

from .PoseTransform import PoseTransform
from .Rigs import Rigs
from .flatten import flatten
from kapture.utils.logging import getLogger
import kapture.utils.computation as computation
from bisect import bisect_left
from copy import deepcopy
import sys
from typing import Union, Dict, List, Optional, Set, Tuple


class Trajectories(Dict[int, Dict[str, PoseTransform]]):
    """
    brief: Trajectories
            trajectories[timestamp][sensor_id] = <PoseTransform>
            or
            trajectories[(timestamp, sensor_id)] = <PoseTransform>

            with <PoseTransform> transforming points from world to device.
    """

    def __init__(self):
        super().__init__()
        self._timestamps_sorted_list = []
        self._first_timestamp = 0
        self._last_timestamp = sys.maxsize

    def __setitem__(self,
                    key: Union[int, Tuple[int, str]],
                    value: Union[Dict[str, PoseTransform], PoseTransform]):
        # enforce type checking
        if isinstance(key, tuple):
            # key is a pair of (timestamp, device_id)
            timestamp = key[0]
            device_id = key[1]
            if not isinstance(timestamp, int):
                raise TypeError('invalid timestamp')
            if not isinstance(device_id, str):
                raise TypeError('invalid device_id')
            if not isinstance(value, PoseTransform):
                raise TypeError('invalid pose')
            self.setdefault(timestamp, {})[device_id] = value
            self._timestamps_sorted_list = []
        elif isinstance(key, int):
            # key is a timestamp
            timestamp = key
            # check type of values?
            if not isinstance(value, dict):
                raise TypeError('invalid value for trajectory timestamp')
            if not all(isinstance(k, str) for k in value.keys()):
                raise TypeError('invalid device_id')
            if not all(isinstance(v, PoseTransform) for v in value.values()):
                raise TypeError('invalid Pose')
            super(Trajectories, self).__setitem__(timestamp, value)
            self._timestamps_sorted_list = []
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def __getitem__(self, key: Union[int, Tuple[int, str]]) -> Union[Dict[str, PoseTransform], PoseTransform]:
        if isinstance(key, tuple):
            # key is a pair of (timestamp, device_id)
            timestamp = key[0]
            device_id = key[1]
            if not isinstance(timestamp, int):
                raise TypeError('invalid timestamp')
            if not isinstance(device_id, str):
                raise TypeError('invalid device_id')
            return super(Trajectories, self).__getitem__(timestamp)[device_id]
        elif isinstance(key, int):
            # key is a timestamp
            return super(Trajectories, self).__getitem__(key)
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def __delitem__(self, key: Union[int, Tuple[int, str]]):
        if isinstance(key, tuple):
            # key is a pair of (timestamp, device_id)
            timestamp = key[0]
            device_id = key[1]
            if not isinstance(timestamp, int):
                raise TypeError('invalid timestamp')
            if not isinstance(device_id, str):
                raise TypeError('invalid device_id')
            super(Trajectories, self).__getitem__(timestamp).__delitem__(device_id)
            if len(super(Trajectories, self).__getitem__(timestamp)) == 0:
                # Cleaning upper level
                super(Trajectories, self).__delitem__(timestamp)
                self._timestamps_sorted_list = []
        elif isinstance(key, int):
            # key is a timestamp
            super(Trajectories, self).__delitem__(key)
            self._timestamps_sorted_list = []
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def timestamps_sorted_list(self) -> List[int]:
        """
        Get the list of timestamps is ascending sorted order
        """
        if len(self._timestamps_sorted_list) == 0:
            # Need to sort
            self._timestamps_sorted_list = sorted(list(self.keys()))
            if len(self._timestamps_sorted_list) > 0:
                self._first_timestamp = self._timestamps_sorted_list[0]
                if len(self._timestamps_sorted_list) > 1:
                    self._last_timestamp = self._timestamps_sorted_list[-1]
        return self._timestamps_sorted_list

    def timestamp_length(self) -> int:
        """
        Compute the trajectory timestamp length, which can be 10, 13 or 16 if these are epoch values, and anything.

        :return: Length of the timestamps as a positive integer, or -1 if it is variable
        """
        timestamps = self.timestamps_sorted_list()
        base_length = computation.num_digits(timestamps[0]) if len(timestamps) > 0 else -1
        indexes = [1, 2, 3, 4, 5, -1, -2, -3, -4] if len(timestamps) > 10 else list(range(1, len(timestamps)))
        for n in indexes:
            length = computation.num_digits(timestamps[n])
            if length != base_length:
                return -1
        return base_length

    def key_pairs(self) -> List[Tuple[int, str]]:
        """
        Returns the list of (timestamp, device_id) contained in trajectories.
        Those pairs can be used to access a single trajectory pose.
        :return: list of (timestamp, device_id)
        """
        return [
            (timestamp, sensor_id)
            for timestamp, sensors in self.items()
            for sensor_id in sensors.keys()
        ]

    @property
    def sensors_ids(self) -> Set[str]:
        """
        :return: the set of unique sensors identifiers in the trajectories
        """
        return set(
            sensor_id
            for timestamp, sensors in self.items()
            for sensor_id in sensors.keys()
        )

    def __contains__(self, key: Union[int, Tuple[int, str]]):
        if isinstance(key, tuple):
            # key is a pair of (timestamp, device_id)
            timestamp = key[0]
            device_id = key[1]
            if not isinstance(timestamp, int):
                raise TypeError('invalid timestamp')
            if not isinstance(device_id, str):
                raise TypeError('invalid device_id')
            return super(Trajectories, self).__contains__(timestamp) and self[timestamp].__contains__(device_id)
        elif isinstance(key, int):
            return super(Trajectories, self).__contains__(key)
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def __repr__(self) -> str:
        # [timestamp, sensor_id] = qw, qx, qy, qz, tx, ty, tz
        lines = [f'[ {timestamp:010}, {sensor_id:5}] = {pose}'
                 for timestamp, sensors in self.items()
                 for sensor_id, pose in sensors.items()]
        return '\n'.join(lines)

    def intermediate_pose(self, timestamp: int, device_id: str, max_interval: int) -> Optional[PoseTransform]:
        """
        Computes an intermediate pose in the trajectory of a device
        The timestamp should be in epoch precision
        i.e. 10, 13, 16 or 19 digits for seconds, milli-seconds, micro-seconds or nano-seconds
        and in the same precision as the trajectories timestamps themselves to work well.
        The timestamps interval should also be not too big (for example one second) to be of good value.
        The max_interval parameter value should be of the same scale as the timestamps.

        :param timestamp: timestamp
        :param device_id: device identifier
        :param max_interval: max interval between the given timestamp and the trajectory timestamps.
        :return: a compute 6D pose if found, None otherwise
        """
        if not isinstance(timestamp, int):
            raise TypeError('invalid timestamp')
        if not isinstance(device_id, str):
            raise TypeError('invalid device_id')
        # In case the pose already exist: just return it
        if self.__contains__(timestamp) and self.__getitem__(timestamp).__contains__(device_id):
            return self.__getitem__(timestamp).__getitem__(device_id)
        timestamps_with_poses_list = self.timestamps_sorted_list()
        # Check if the pose is out of bounds
        if timestamp <= self._first_timestamp or timestamp >= self._last_timestamp:
            return None
        # Find closest timestamps before and after
        next_position = bisect_left(timestamps_with_poses_list, timestamp)
        low_position = next_position - 1
        previous_ts = timestamps_with_poses_list[low_position]
        next_ts = timestamps_with_poses_list[next_position]
        # We should have found the two closest timestamps
        # Check there is a pose in the time interval for the device
        while timestamp - previous_ts <= max_interval and not self.__getitem__(previous_ts).__contains__(device_id):
            # Search backward for a timestamp with this device
            low_position -= 1
            if low_position >= 0:
                previous_ts = timestamps_with_poses_list[low_position]
            else:
                # We have reached the begin of the list without solution
                return None
        # Check the interval for the previous timestamp
        if timestamp - previous_ts > max_interval:
            # We are to far in the past
            return None
        while next_ts - timestamp <= max_interval and not self.__getitem__(next_ts).__contains__(device_id):
            # Search backward for a timestamp with this device
            next_position += 1
            if next_position < len(timestamps_with_poses_list):
                next_ts = timestamps_with_poses_list[next_position]
            else:
                # We have reached the end of the list without solution
                return None
        # Check the interval for the next timestamp
        if next_ts - timestamp > max_interval:
            # We are to far in the future
            return None
        previous_pose = self.__getitem__(previous_ts).__getitem__(device_id)
        next_pose = self.__getitem__(next_ts).__getitem__(device_id)
        return compute_intermediate_pose(timestamp, previous_ts, previous_pose, next_ts, next_pose)


def rigs_remove(trajectories: Trajectories, rigs: Rigs) -> Trajectories:
    """
    Returns a new trajectories based on the given one,
    but where rigs poses are replaced by the poses of every sensors.
    This is useful for formats that does not have the rig notion.

    :param trajectories: input Trajectories where the rigs has to be replaced
    :param rigs: input Rigs that defines the rigs/sensors relationship.
    :return: output Trajectories where the rigs has been replaced.
    """
    assert isinstance(rigs, Rigs)
    assert isinstance(trajectories, Trajectories)
    new_trajectories = deepcopy(trajectories)
    rigs_remove_inplace(new_trajectories, rigs)
    return new_trajectories


def rigs_remove_inplace(trajectories: Trajectories, rigs: Rigs, max_depth: int = 10):
    """
    Removes rigs poses and replaces them by the poses of every sensors in it.
    The operation is performed inplace, and modifies trajectories.
    This is useful for formats that does not have the rig notion.

    :param trajectories: input/output Trajectories where the rigs has to be replaced
    :param rigs: input Rigs that defines the rigs/sensors relationship.
    :param max_depth: maximum nested rig depth.
    """
    assert isinstance(rigs, Rigs)
    assert isinstance(trajectories, Trajectories)
    # collect all poses of rigs in trajectories
    for iteration in range(max_depth):
        # repeat the operation while there is so rig remaining (nested rigs)
        jobs = [(timestamp, rig_id, pose_rig_from_world)
                for timestamp, poses_for_timestamp in trajectories.items()
                for rig_id, pose_rig_from_world in poses_for_timestamp.items()
                if rig_id in rigs]

        if len(jobs) == 0:
            # we are done
            break

        getLogger().debug(f'rigs_remove {len(jobs)} jobs at depth {iteration}')
        for timestamp, rig_id, pose_rig_from_world in tqdm(jobs, disable=getLogger().level >= logging.CRITICAL):
            for device_id, pose_device_from_rig in rigs[rig_id].items():
                pose_cam_from_world = PoseTransform.compose([pose_device_from_rig, pose_rig_from_world])
                trajectories.setdefault(timestamp, {})[device_id] = pose_cam_from_world
            del trajectories[timestamp][rig_id]

    for timestamp in trajectories.keys():
        if len(trajectories[timestamp]) == 0:
            del trajectories[timestamp]
    # Do not clear, the rigs : so easy to do outside, and so easy


def rigs_recover(
        trajectories: Trajectories,
        rigs: Rigs,
        master_sensors: Optional[List[str]] = None
) -> Trajectories:
    """
    Returns a new trajectories based on the given one,
    but where sensor poses are replaced by the rig pose, each time the sensor is registered in a rig.
    The rig configuration must be given.
    Warning: in some configuration, it may leads to inconsistent results. See code for details.

    :param trajectories: input Trajectories.
    :param rigs: input Rigs configuration.
    :param master_sensors: input If given, only compute rig poses for the given sensors.
    :return: a new Trajectories.
    """
    new_trajectories = deepcopy(trajectories)
    rigs_recover_inplace(new_trajectories, rigs, master_sensors)
    return new_trajectories


def rigs_recover_inplace(
        trajectories: Trajectories,
        rigs: Rigs,
        master_sensors: Optional[List[str]] = None,
        max_depth: int = 10
):
    """
    Updates the given trajectories by replacing sensor poses by the rig poses,
    each time the sensor is registered in a rig.
    The rig configuration must be given.
    Warning: in some configuration, it may leads to inconsistent results. See code for details.

    :param trajectories: input/output Trajectories.
    :param rigs: input Rigs configuration.
    :param master_sensors: input If given, only compute rig poses for the given sensors.
    :param max_depth: maximum nested rig depth.
    """

    # sensor_id -> rig_id, pose_rig_from_sensor
    reverse_rig_dict = {
        sensor_id: (rig_id, rigs[rig_id, sensor_id].inverse())
        for rig_id, sensor_id in rigs.key_pairs()
    }

    for iteration in range(max_depth):
        # do the replacement while there is sensor_id in trajectories, that can be converted to rig_id
        jobs = [(timestamp, sensor_id, pose_sensor_from_world)
                for timestamp, sensor_id, pose_sensor_from_world in flatten(trajectories, is_sorted=True)
                if sensor_id in reverse_rig_dict]

        if len(jobs) == 0:
            break

        for timestamp, sensor_id, pose_sensor_from_world in jobs:
            # if the sensor is part of a rig, set the pose of the rig,
            # instead of the pose of the sensor
            rig_id, pose_rig_from_sensor = reverse_rig_dict[sensor_id]
            pose_sensor_from_world = trajectories[timestamp].pop(sensor_id)

            # only compute rig poses for master_sensors if any.
            if master_sensors is not None and sensor_id not in master_sensors:
                continue

            # skip if rig pose already recovered.
            if rig_id in trajectories[timestamp]:
                continue

            # warning: if multiple sensors can be used to infer the rig pose
            # (eg. no master sensor, or multiple master sensors), only one sensor is used to compute the rig pose.
            # Since we use is_sorted=True, the selected sensor should be consistent, when present. But if for some
            # timestamp, the usual first sensor is missing, then, another sensor is used as reference and
            # if the sensors dos not actually use the rig calibration (no rigid transform between sensors),
            # it may end up to inconsistent results.
            pose_rig_from_world = PoseTransform.compose([pose_rig_from_sensor, pose_sensor_from_world])
            trajectories[timestamp, rig_id] = pose_rig_from_world


def compute_intermediate_pose(timestamp: int,
                              low_ts: int, low_p: PoseTransform,
                              up_ts: int, up_p: PoseTransform) -> PoseTransform:
    """
    Compute an intermediate pose between two poses.
    It does not come from the recorded data, but is purely computed by interpolation based on given poses.
    We suppose that we move at a regular speed.

    :param timestamp: the timestamp at which time to compute the pose
    :param low_ts: the first timestamp
    :param low_p: the first pose
    :param up_ts: the second timestamp
    :param up_p: the second pose
    :return: the computed pose
    """
    rotation = quaternion.slerp(low_p.r, up_p.r, low_ts, up_ts, timestamp)
    # translation = t0 + (ts-ts0)/(ts1-ts0) * (t1 - t0)
    translation = [low_p.t[0] + (timestamp - low_ts) / (up_ts - low_ts) * (up_p.t[0] - low_p.t[0]),
                   low_p.t[1] + (timestamp - low_ts) / (up_ts - low_ts) * (up_p.t[1] - low_p.t[1]),
                   low_p.t[2] + (timestamp - low_ts) / (up_ts - low_ts) * (up_p.t[2] - low_p.t[2])]
    return PoseTransform(rotation, translation)



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script imports data from a COLMAP database and/or reconstruction files

It will assume that if all rotation parameters are null, there is no prior rotation, and if all translation parameters
are null, there is no prior translation (using null values instead of zeros)
"""

import argparse
import logging
import os
import sys

import path_to_kapture  # enables import kapture  # noqa: F401
import kapture
import kapture.utils.logging
import kapture.io.csv
from kapture.io.records import TransferAction
from kapture.converter.colmap.import_colmap import import_colmap

logger = logging.getLogger('colmap')


def colmap_command_line() -> None:
    """
    Do the colmap import to kapture using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(description='imports data from colmap database and/or text files.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete kapture if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-db', '--database',
                        help='path to COLMAP database file.')
    parser.add_argument('-txt', '--reconstruction',
                        help='path to COLMAP reconstruction triplet text file.')
    # images
    parser.add_argument('-im', '--images', default=None,
                        help='path to images directory.')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.link_absolute,
                        help=f'How to import images [link_absolute],'
                             f'choose among: {", ".join(a.name for a in TransferAction)}')
    parser.add_argument('-rig', '--rig',
                        help='path to COLMAP rig json file.')
    parser.add_argument('-o', '-k', '--kapture', required=True,
                        help='output directory where to save kapture files.')
    parser.add_argument('-kpt', '--keypoints-type', default='SIFT', help='kapture keypoints type.')
    parser.add_argument('-desc', '--descriptors-type', default='SIFT', help='kapture descriptors type.')
    parser.add_argument('--skip_reconstruction', action='store_true', default=False,
                        help='skip the import of the kapture/reconstruction,'
                             'ie. Keypoints, Descriptors, Matches, Points3d, Observations.')
    parser.add_argument('--no_geometric_filtering', action='store_true', default=False,
                        help='don\'t remove matches failing geometric verification.')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # for debug, let kapture express itself.
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    logger.debug(f'{sys.argv[0]} \\\n' + '  \\\n'.join(
        '--{:20} {:100}'.format(k, str(v))
        for k, v in vars(args).items()))

    if args.image_transfer == TransferAction.root_link:
        logger.fatal('Root linking the images is not supported yet.')
        return
    logger.debug('cleaning output directory')
    os.makedirs(args.kapture, exist_ok=True)
    kapture.io.structure.delete_existing_kapture_files(args.kapture, force_erase=args.force)

    logger.info('importing colmap ...')
    kapture_data = import_colmap(args.kapture,
                                 args.database,
                                 args.reconstruction,
                                 args.images,
                                 args.rig,
                                 args.keypoints_type,
                                 args.descriptors_type,
                                 args.no_geometric_filtering,
                                 args.skip_reconstruction,
                                 args.force,
                                 args.image_transfer if args.images is not None else TransferAction.skip
                                 )

    logger.info('saving to kapture  ...')
    logger.debug(f'\t"{args.kapture}"')
    kapture.io.csv.kapture_to_dir(args.kapture, kapture_data)


if __name__ == '__main__':
    colmap_command_line()



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script imports a RobotCar dataset to the kapture format.


    The RobotCar Seasons dataset uses a subset of the images provided in the RobotCar
    dataset. It uses images taken under a single reference condition (`overcast-reference`),
    captured at 49 different non-overlapping locations, to represent the scene. For this reference
    condition, the dataset provides a reference 3D model reconstructed using COLMAP [1,2]. The
    3D model consequently defines a set of 6DOF reference poses for the database images.
    In addition, query images taken under different conditions (`dawn`, `dusk`, `night`,
    `night-rain`, `overcast-summer`, `overcast-winter`, `rain`, `snow`, `sun`) at the 49
    locations are provided. For these query images, the reference 6DOF poses will not be
    released.

    T. Sattler, W. Maddern, C. Toft, A. Torii, L. Hammarstrand, E. Stenborg, D. Safari, M. Okutomi, M. Pollefeys,
     J. Sivic, F. Kahl, T. Pajdla.
    Benchmarking 6DOF Outdoor Visual Localization in Changing Conditions.
    Conference on Computer Vision and Pattern Recognition (CVPR) 2018

    The dataset is based on the RobotCar dataset described in this paper:

    W. Maddern, G. Pascoe, C. Linegar, and P. Newman.
    1 Year, 1000km: The Oxford RobotCar Dataset.
    International Journal of Robotics Research (IJRR), 36(1):3–15, 2017

    https://www.visuallocalization.net/datasets/


    Importing:
    - convert all 49 individual colmap models
    - restore rig and timestamps
      - image names: png to jpg
    - convert colmap reference DB to kapture -> needed?
    - merge -> needed?

"""

import argparse
import logging
import numpy as np
import os
import os.path as path
import quaternion
import re
from typing import Dict
# kapture
import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.csv import kapture_to_dir
from kapture.io.records import TransferAction, import_record_data_from_dir_auto
from kapture.converter.colmap.import_colmap import import_colmap

logger = logging.getLogger('RobotCar_Seasons')
LOCAL_FEATURE_TYPE = 'SIFT'


def import_robotcar_cameras(intrinsics_dir_path: str) -> kapture.Sensors:
    """
    Read and convert intrinsics files

    :param intrinsics_dir_path:
    :return: kapture.cameras
    """
    cameras = kapture.Sensors()
    for root, dirs, files in os.walk(intrinsics_dir_path):
        for intrinsic_filename in files:
            (camera_id, _) = intrinsic_filename.split('_')
            intrinsic_file = open(path.join(intrinsics_dir_path, intrinsic_filename), 'r')
            (_, fx) = intrinsic_file.readline().split()
            (_, fy) = intrinsic_file.readline().split()
            (_, cx) = intrinsic_file.readline().split()
            (_, cy) = intrinsic_file.readline().split()
            intrinsic_file.close()
            # w, h, fx, fy, cx, cy
            model = kapture.CameraType.PINHOLE
            model_params = [1024, 1024, fx, fy, cx, cy]
            cameras[camera_id] = kapture.Camera(model, model_params)

    return cameras


def import_robotcar_rig(extrinsics_dir_path: str) -> kapture.Rigs:
    """
    Read extrinsics files and convert to kapture rigs

    :param extrinsics_dir_path: path to directory with extrinsics
    :return: kapture.Rigs object
    """
    # From dataset documentation:
    #   The extrinsic parameters of the three cameras on the car with respect to the car itself.
    #   The extrinsics are provided as a 4x4 matrix
    #     [R c]
    #     [0 1]
    #   where R is a rotation matrix that maps from camera to world coordinates and c is the position of the camera in
    #   world coordinates. The entries of the matrix are separated by ,.

    rigs = kapture.Rigs()
    for root, dirs, files in os.walk(extrinsics_dir_path):
        for extrinsics_filepath in files:
            (camera_id, _) = extrinsics_filepath.split('_')
            extrinsic_file = open(path.join(extrinsics_dir_path, extrinsics_filepath), 'r')
            with extrinsic_file as f:
                m = np.array([[float(num) for num in line.split(',')] for line in f])
            rotation_matrix = m[0:3, 0:3]
            position = m[0:3, 3:4]
            pose_transform = kapture.PoseTransform(t=position, r=quaternion.from_rotation_matrix(rotation_matrix))
            # kapture rig format is "rig to sensor"
            rigs['car', camera_id] = pose_transform.inverse()

    return rigs


def import_robotcar_colmap_location(robotcar_path: str,  # noqa: C901: function a bit long but not too complex
                                    colmap_reconstruction_fullpath: path,
                                    kapture_path: str,
                                    rigs: kapture.Rigs,
                                    skip_reconstruction: bool) -> kapture.Kapture:
    """
    Import robotcar data for one location from colmap reconstruction

    :param robotcar_path: path to the robotcar top directory
    :param colmap_reconstruction_fullpath: path to the colmap reconstruction directory
    :param kapture_path: path to the kapture top directory
    :param rigs: kapture rigs to modify
    :param skip_reconstruction: if True, will not add the reconstruction
    :return: a kapture object
    """

    # First, import Colmap reconstruction for given location
    kapture_data = import_colmap(kapture_dir_path=kapture_path,
                                 colmap_reconstruction_dir_path=colmap_reconstruction_fullpath,
                                 colmap_images_dir_path=path.join(robotcar_path, "images"),
                                 keypoints_type=LOCAL_FEATURE_TYPE,
                                 descriptors_type=LOCAL_FEATURE_TYPE,
                                 skip_reconstruction=skip_reconstruction,
                                 images_import_strategy=TransferAction.skip)  # since filenames are incorrect

    # Post processing:
    # - use correct names for cameras
    # - model was built with PNG files, but we have JPG
    # - recover proper timestamps
    # - recover rig

    # Fix sensors.txt
    camera_mapping = {
        'cam_00001': 'left',
        'cam_00002': 'rear',
        'cam_00003': 'right'
    }
    new_cameras = kapture.Sensors()
    for cam_id in kapture_data.sensors:
        new_cameras[camera_mapping[cam_id]] = kapture_data.sensors[cam_id]
    kapture_data.sensors = new_cameras

    if not skip_reconstruction:
        # Fix keypoints
        # Need to rename .png.kpt to .jpg.kpt files and that's all
        for root, dirs, files in os.walk(kapture_path):
            for file in files:
                if file.endswith('.png.kpt'):
                    os.rename(path.join(root, file), path.join(root, file.replace(".png.kpt", ".jpg.kpt")))

        # observations.txt: png -> jpg
        new_observations = kapture.Observations()
        for point3d_idx in kapture_data.observations:
            for image_path, keypoint_id in kapture_data.observations[point3d_idx, LOCAL_FEATURE_TYPE]:
                new_observations.add(point3d_idx, LOCAL_FEATURE_TYPE,
                                     image_path.replace(".png", ".jpg"), int(keypoint_id))
        kapture_data.observations = new_observations

    # records_camera.txt
    # timestamps, png->jpg
    new_records_camera = kapture.RecordsCamera()
    records_camera_pattern = re.compile(r'.*/(?P<timestamp>\d+)\.png')
    ts_mapping = {}
    for ts, shot in kapture_data.records_camera.items():
        for cam_id, image_path in shot.items():
            matches = records_camera_pattern.match(image_path)
            if not matches:
                continue
            matches = matches.groupdict()
            timestamp = int(matches['timestamp'])
            ts_mapping[ts] = timestamp
            new_path = image_path.replace(".png", ".jpg")
            new_records_camera[timestamp, camera_mapping[cam_id]] = new_path
    kapture_data.records_camera = new_records_camera

    # trajectories.txt
    new_trajectories = kapture.Trajectories()
    # First recover timestamps and camera names
    for ts, sensor_id in sorted(kapture_data.trajectories.key_pairs()):
        new_trajectories[ts_mapping[ts], camera_mapping[sensor_id]] = kapture_data.trajectories[ts, sensor_id]

    kapture_data.trajectories = new_trajectories
    kapture_data.rigs = rigs

    return kapture_data


def read_robotcar_v2_train(robotcar_path: str) -> Dict[str, kapture.PoseTransform]:
    """
    Read the robot car v2 train data file

    :param robotcar_path: path to the data file
    :return: train data
    """
    with(open(path.join(robotcar_path, 'robotcar_v2_train.txt'), 'r')) as f:
        lines = f.readlines()
        # remove empty lines
        lines = (l2 for l2 in lines if l2.strip())
        # trim trailing EOL
        lines = (l3.rstrip("\n\r") for l3 in lines)
        # split space
        lines = (l4.split() for l4 in lines)
        lines = list(lines)
        train_data = {table[0].replace(".png", ".jpg"):
                      kapture.PoseTransform(r=quaternion.from_rotation_matrix([[float(v) for v in table[1:4]],
                                                                               [float(v) for v in table[5:8]],
                                                                               [float(v) for v in table[9:12]]]),
                                            t=[float(v) for v in [table[4], table[8], table[12]]]).inverse()
                      for table in lines}
    return train_data


def import_robotcar_seasons(robotcar_path: str,  # noqa: C901: function a bit long but not too complex
                            kapture_path: str,
                            force_overwrite_existing: bool = False,
                            images_import_method: TransferAction = TransferAction.skip,
                            import_feature_db: bool = False,
                            skip_reconstruction: bool = False,
                            rig_collapse: bool = False,
                            use_colmap_intrinsics: bool = False,
                            import_v1: bool = False) -> None:
    """
    Read the RobotCar Seasons data, creates several kaptures with training and query data.

    :param robotcar_path: path to the robotcar top directory
    :param kapture_path: path to the kapture top directory
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    :param images_import_method: choose how to import actual image files.
    :param import_feature_db: if True, will import the features from the database
    :param skip_reconstruction: if True, will skip the reconstruction part from the training data
    :param rig_collapse: if True, will collapse the rig
    :param use_colmap_intrinsics: if True, will use the colmap intrinsics
    :param import_v1: if True, will use the version 1 of the format
    """

    os.makedirs(kapture_path, exist_ok=True)

    cameras = import_robotcar_cameras(path.join(robotcar_path, 'intrinsics'))
    rigs = import_robotcar_rig(path.join(robotcar_path, 'extrinsics'))

    logger.info("Importing test data")
    # Test data
    image_pattern = re.compile(r'(?P<condition>.+)/(?P<camera>\w+)/(?P<timestamp>\d+)\.jpg')
    queries_path = path.join(robotcar_path, '3D-models', 'individual', 'queries_per_location')
    kapture_imported_query = {}
    for root, dirs, files in os.walk(queries_path):
        for query_file in files:
            records_camera = kapture.RecordsCamera()
            # Get list of query images
            with open(path.join(queries_path, query_file)) as f:
                for line in f:
                    matches = image_pattern.match(line)
                    image_path = line.strip()
                    if not matches:
                        logger.warning(f"Error matching line in {image_path}")
                        continue
                    matches = matches.groupdict()
                    timestamp = int(matches['timestamp'])
                    camera = str(matches['camera'])
                    # condition = str(matches['condition']) : not used ?
                    records_camera[timestamp, camera] = image_path

                (query_name, _) = query_file.split('.')
                kapture_test = kapture.Kapture(sensors=cameras, rigs=rigs, records_camera=records_camera)
                kapture_imported_query[int(query_name.split('_')[-1])] = kapture_test

    # Training data
    logger.info("Importing training data")
    colmap_reconstructions_path = path.join(robotcar_path, '3D-models', 'individual', 'colmap_reconstructions')
    kapture_imported_training = {}
    for root, dirs, files in os.walk(colmap_reconstructions_path):
        for colmap_reconstruction in dirs:
            (loc_id, _) = colmap_reconstruction.split('_')
            kapture_reconstruction_dir = path.join(kapture_path, f"{int(loc_id):02d}", "mapping")
            delete_existing_kapture_files(kapture_reconstruction_dir, force_overwrite_existing)
            logger.info(f'Converting reconstruction {loc_id} to kapture  ...')
            kapture_reconstruction_data = import_robotcar_colmap_location(
                robotcar_path,
                path.join(colmap_reconstructions_path, colmap_reconstruction),
                kapture_reconstruction_dir,
                rigs,
                skip_reconstruction
            )
            # replace intrinsics with the ones found in the text files
            if not use_colmap_intrinsics:
                kapture_reconstruction_data.sensors = cameras
            kapture_imported_training[int(loc_id)] = kapture_reconstruction_data

    if not import_v1:
        _import_robotcar_v2_train(robotcar_path, kapture_imported_query, kapture_imported_training, image_pattern)

    # apply rig collapse
    if rig_collapse:
        logger.info('replacing camera poses with rig poses.')
        for kapture_mapping in kapture_imported_training.values():
            kapture.rigs_recover_inplace(kapture_mapping.trajectories, rigs, ['rear'])

    # IO operations
    robotcar_image_path = path.join(robotcar_path, "images")
    for loc_id, kapture_query in kapture_imported_query.items():
        loc_id_str = f"{loc_id:02d}"
        logger.info(f'writing test data: {loc_id_str}')
        kapture_test_dir = path.join(kapture_path, loc_id_str, "query")
        delete_existing_kapture_files(kapture_test_dir, force_overwrite_existing)
        if not kapture_query.records_camera:  # all images were removed
            continue
        kapture_to_dir(kapture_test_dir, kapture_query)
        query_images = [f for _, _, f in kapture.flatten(kapture_query.records_camera)]
        import_record_data_from_dir_auto(robotcar_image_path, kapture_test_dir,
                                         query_images, images_import_method)

    for loc_id, kapture_mapping in kapture_imported_training.items():
        loc_id_str = f"{loc_id:02d}"
        logger.info(f'writing mapping data: {loc_id_str}')
        kapture_reconstruction_dir = path.join(kapture_path, f"{loc_id:02d}", "mapping")
        kapture_to_dir(kapture_reconstruction_dir, kapture_mapping)
        mapping_images = [f for _, _, f in kapture.flatten(kapture_mapping.records_camera)]
        import_record_data_from_dir_auto(robotcar_image_path, kapture_reconstruction_dir,
                                         mapping_images, images_import_method)

    if import_feature_db:
        _import_colmap_overcast_reference(robotcar_path, kapture_path, force_overwrite_existing)


def _import_robotcar_v2_train(robotcar_path, kapture_imported_query, kapture_imported_training, image_pattern):
    queries_per_location = {image_name: (ts, cam_id, loc_id)
                            for loc_id, kapture_test in kapture_imported_query.items()
                            for ts, cam_id, image_name in kapture.flatten(kapture_test.records_camera)}
    # read robotcar_v2_train.txt
    v2_train_data = read_robotcar_v2_train(robotcar_path)
    for image_name, pose in v2_train_data.items():
        ts, cam_id, loc_id = queries_per_location[image_name]
        assert cam_id == 'rear'
        kapture_imported_training[loc_id].records_camera[ts, cam_id] = image_name
        kapture_imported_training[loc_id].trajectories[ts, cam_id] = pose
        matches = image_pattern.match(image_name)
        if not matches:
            logger.warning(f"Error matching line in {image_name}")
            continue
        matches = matches.groupdict()
        condition = str(matches['condition'])
        timestamp = str(matches['timestamp'])
        # added left and right images in records_camera
        left_image_name = condition + '/' + 'left' + '/' + timestamp + '.jpg'
        right_image_name = condition + '/' + 'right' + '/' + timestamp + '.jpg'
        kapture_imported_training[loc_id].records_camera[ts, 'left'] = left_image_name
        kapture_imported_training[loc_id].records_camera[ts, 'right'] = right_image_name

        # remove entries from query
        del kapture_imported_query[loc_id].records_camera[ts][cam_id]
        del kapture_imported_query[loc_id].records_camera[ts]['left']
        del kapture_imported_query[loc_id].records_camera[ts]['right']
        del kapture_imported_query[loc_id].records_camera[ts]
    # all remaining query images are kept; reading robotcar_v2_test.txt is not necessary


def _import_colmap_overcast_reference(robotcar_path, kapture_path, force_overwrite_existing):
    # Convert Colmap reference DB to kapture
    kapture_train_dir = path.join(kapture_path, "mapping")
    colmap_db_path = path.join(robotcar_path, "3D-models/overcast-reference.db")
    if path.exists(colmap_db_path):
        delete_existing_kapture_files(kapture_train_dir, force_overwrite_existing)
        kapture_train_data = import_colmap(
            kapture_dir_path=kapture_train_dir,
            colmap_database_filepath=colmap_db_path,
            colmap_reconstruction_dir_path='',
            colmap_images_dir_path=path.join(robotcar_path, "images"),
            keypoints_type=LOCAL_FEATURE_TYPE,
            descriptors_type=LOCAL_FEATURE_TYPE,
            no_geometric_filtering=True,
            force_overwrite_existing=force_overwrite_existing,
            images_import_strategy=TransferAction.skip
        )
        logger.info(f'saving feature DB to kapture {kapture_train_dir}  ...')
        kapture_to_dir(kapture_train_dir, kapture_train_data)
    else:
        logger.warning(f'Colmap feature DB {colmap_db_path} does not exist... skipping.')


def import_robotcar_seasons_command_line() -> None:
    """
    Do the RobotCar dataset import to kapture using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(
        description='imports RobotCar Seasons dataset to the kapture format. '
                    'This script will create one kapture dataset'
                    ' for each of the 49 locations, based on the colmap reconstructions')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-i', '--input', required=True, help='input path to RobotCar-Seasons root folder')
    parser.add_argument('-o', '--output', required=True, help='kapture output directory')
    parser.add_argument('--import_feature_db', required=False, action='store_true', default=False,
                        help='also convert colmap feature database to kapture format')
    parser.add_argument('--skip_reconstruction', required=False, action='store_true', default=False,
                        help='do not import reconstruction data (3d points, keypoints and observations')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.root_link,
                        help=(f'How to import images [root_link], '
                              f'choose among: {", ".join(a.name for a in TransferAction)}'))
    parser.add_argument('--rig_collapse', action='store_true', default=False,
                        help='Replace camera poses with rig poses.')
    parser.add_argument('--use_colmap_intrinsics', action='store_true', default=False,
                        help=('For mapping images, use intrinsics from the reconstruction (SIMPLE_RADIAL)'
                              ' instead of the intrinsics text files (PINHOLE)'))
    parser.add_argument('--v1', action='store_true', default=False,
                        help='Export RobotCar_Seasons v1 instead of v2.')
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    import_robotcar_seasons(args.input, args.output, args.force, args.image_transfer,
                            args.import_feature_db, args.skip_reconstruction, args.rig_collapse,
                            args.use_colmap_intrinsics, args.v1)


if __name__ == '__main__':
    import_robotcar_seasons_command_line()



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to import a 12scenes model into a kapture.
The RGB-D Dataset 12-Scenes data structure is defined here:
    http://graphics.stanford.edu/projects/reloc/

Each sequence contains:
    Color frames (frame-XXXXXX.color.jpg): RGB, 24-bit, JPG
    Depth frames (frame-XXXXXX.depth.png): depth (mm), 16-bit, PNG (invalid depth is set to 0)
    Camera poses (frame-XXXXXX.pose.txt): camera-to-world
    Camera calibration (info.txt): color and depth camera intrinsics and extrinsics.
    Note that these are the default intrinsics and we did not perform any calibration.
"""

import argparse
import logging
import os
import os.path as path
import re
import numpy as np
import quaternion
from typing import Optional
from PIL import Image
from tqdm import tqdm
from scipy import ndimage
# kapture
import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.csv import kapture_to_dir
import kapture.io.features
from kapture.io.records import TransferAction, import_record_data_from_dir_auto
from kapture.utils.paths import path_secure

logger = logging.getLogger('12scenes')
MODEL = kapture.CameraType.RADIAL

POSE_SUFFIX = 'pose'
RGB_SUFFIX = 'color'
DEPTH_SUFFIX = 'depth'
RGB_SENSOR_ID = 'ipad_camera'
DEPTH_SENSOR_ID = 'structure_io_depth_camera'
REG_DEPTH_SENSOR_ID = 'structure_io_depth_camera_reg'
RGBD_SENSOR_ID = 'rgbd_rig'


def get_camera_matrix(fx: float, fy: float, cx: float, cy: float):
    return np.array([[fx, 0, cx],
                     [0, fy, cy],
                     [0, 0, 1]])


def get_K(camera_type: kapture.CameraType, camera_params):
    if camera_type == kapture.CameraType.UNKNOWN_CAMERA:
        width = camera_params[0]
        height = camera_params[1]
        focal = 1.2 * max(width, height)
        return get_camera_matrix(focal, focal, camera_params[0] / 2, camera_params[1] / 2)
    elif camera_type in [kapture.CameraType.PINHOLE, kapture.CameraType.OPENCV,
                         kapture.CameraType.OPENCV_FISHEYE, kapture.CameraType.FULL_OPENCV,
                         kapture.CameraType.FOV, kapture.CameraType.THIN_PRISM_FISHEYE]:
        return get_camera_matrix(camera_params[2], camera_params[3], camera_params[4], camera_params[5])
    else:
        return get_camera_matrix(camera_params[2], camera_params[3], camera_params[4], camera_params[4])


def dilate(x, k):
    m, n = x.shape
    y = np.empty_like(x)
    for i in range(0, m):
        for j in range(0, n):
            currmax = x[i, j]
            for ii in range(int(max(0, i-k/2)), int(min(m, i+k/2+1))):
                for jj in range(int(max(0, j-k/2)), int(min(n, j+k/2+1))):
                    elt = x[ii, jj]
                    if elt > currmax:
                        currmax = elt
            y[i, j] = currmax
    return y


def register_depth(Kdepth, Kcolor, Rt, depth, width_color, height_color):
    reg_depth = np.zeros((height_color, width_color), dtype=np.float32)
    y, x = np.meshgrid(np.arange(depth.shape[0]), np.arange(depth.shape[1]), indexing='ij')
    x = x.reshape(1, -1)
    y = y.reshape(1, -1)
    z = depth.reshape(1, -1)
    x = (x - Kdepth[0, 2]) / Kdepth[0, 0]
    y = (y - Kdepth[1, 2]) / Kdepth[1, 1]
    pts = np.vstack((x * z, y * z, z))
    pts = Rt[:3, :3] @ pts + Rt[:3, 3:]
    pts = Kcolor @ pts
    px = np.round(pts[0, :] / pts[2, :])
    py = np.round(pts[1, :] / pts[2, :])
    mask = (px >= 0) * (py >= 0) * (px < width_color) * (py < height_color)
    reg_depth[py[mask].astype(int), px[mask].astype(int)] = pts[2, mask]

    reg_depth = ndimage.grey_dilation(reg_depth, size=(3, 3))

    return reg_depth


def import_12scenes(d12scenes_path: str,
                    kapture_dir_path: str,
                    force_overwrite_existing: bool = False,
                    images_import_method: TransferAction = TransferAction.skip,
                    partition: Optional[str] = None
                    ) -> None:
    """
    Imports RGB-D Dataset 12-Scenes dataset and save them as kapture.

    :param d12scenes_path: path to the 12scenes sequence root path
    :param kapture_dir_path: path to kapture top directory
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    :param images_import_method: choose how to import actual image files.
    :param partition: if specified = 'mapping' or 'query'.
    """
    os.makedirs(kapture_dir_path, exist_ok=True)
    delete_existing_kapture_files(kapture_dir_path, force_erase=force_overwrite_existing)

    logger.info('loading all content ...')

    d7s_filename_re = re.compile(r'frame-(?P<frame_id>\d{6})\.(?P<suffix>\w*)\.(?P<ext>\w*)')

    # populate all relevant files
    d12images_path = os.path.join(d12scenes_path, 'data')
    d7s_filenames = (path_secure(path.relpath(path.join(dp, fn), d12images_path))
                     for dp, _, fs in os.walk(d12images_path) for fn in fs)

    logger.info('populating 12-scenes files ...')
    d7s_filenames = {filename: d7s_filename_re.search(filename).groupdict()
                     for filename in sorted(d7s_filenames)
                     if d7s_filename_re.search(filename)}

    # reorg as shot[seq, id] = {color: , depth: , pose: , ...}
    shots = {}
    for timestamp, (filename, file_attribs) in enumerate(d7s_filenames.items()):
        shot_id = int(file_attribs['frame_id'])
        shots.setdefault(shot_id, {})[file_attribs['suffix']] = filename

    # fake timestamps
    for timestamp, shot_id in enumerate(shots):
        shots[shot_id]['timestamp'] = timestamp

    # if given, filter partition
    if partition is not None:
        # read the authors split file
        partition_filepath = path.join(d12scenes_path, 'split.txt')
        if not path.isfile(partition_filepath):
            raise FileNotFoundError(f'partition file is missing: {partition_filepath}.')

        with open(partition_filepath, 'rt') as file:
            # note from dsac++; the first sequence is used for testing, everything else for training
            d7s_split_exp = r'^sequence(?P<sequence>\d+) \[frames=(?P<count>\d+)\]  \[start=(?P<start_frame>\d+) ;' \
                            r' end=(?P<end_frame>\d+)\]$'
            d7s_split_re = re.compile(d7s_split_exp)
            split_sequences = [re.match(d7s_split_re, line) for line in file.readlines()]
            if len(split_sequences) < 1 or not split_sequences[0]:
                raise ValueError('failed to parse split.txt file')
            test_split = (int(split_sequences[0].group('start_frame')), int(split_sequences[0].group('end_frame')))

            # filter out
            if partition == "query":
                shots = {frame: shot
                         for frame, shot in shots.items()
                         if test_split[0] <= frame <= test_split[1]
                         }
            elif partition == "mapping":
                shots = {frame: shot
                         for frame, shot in shots.items()
                         if frame < test_split[0] or frame > test_split[1]
                         }
            else:
                raise ValueError('invalid partition name')

    if len(shots) == 0:
        raise FileNotFoundError('no file found: make sure the path to 12scenes sequence is valid.')

    # eg. shots['000000'] =
    #       {
    #           'color': 'seq-01/frame-000000.color.jpg',
    #           'depth': 'seq-01/frame-000000.depth.png',
    #           'pose': 'seq-01/frame-000000.pose.txt',
    #           'timestamp': 0}

    # images + depth maps
    logger.info('populating image and depth maps files ...')
    snapshots = kapture.RecordsCamera()
    depth_maps = kapture.RecordsDepth()
    for shot in shots.values():
        snapshots[shot['timestamp'], RGB_SENSOR_ID] = shot['color']
        kapture_depth_map_filename = shot['depth'][:-len('.png')]  # kapture depth files are not png
        depth_maps[shot['timestamp'], DEPTH_SENSOR_ID] = kapture_depth_map_filename
        kapture_registered_depth_map_filename = shot['depth'][:-len('.png')] + '.reg'  # kapture depth files are not png
        depth_maps[shot['timestamp'], REG_DEPTH_SENSOR_ID] = kapture_registered_depth_map_filename

    # poses
    logger.info('import poses files ...')
    trajectories = kapture.Trajectories()
    for shot in shots.values():
        pose_filepath = path.join(d12images_path, shot['pose'])
        pose_mat = np.loadtxt(pose_filepath)  # camera-to-world, 4×4 matrix in homogeneous coordinates
        with open(pose_filepath, 'r') as file:
            if 'INF' in file.read():
                timestamp = shot['timestamp']
                image_name = shot['color']
                logger.debug(f'ts={timestamp}, name={image_name}: ignored inf pose')
                continue
        rotation_mat = pose_mat[0:3, 0:3]
        position_vec = pose_mat[0:3, 3]
        rotation_quat = quaternion.from_rotation_matrix(rotation_mat)
        pose_world_from_cam = kapture.PoseTransform(r=rotation_quat, t=position_vec)
        pose_cam_from_world = pose_world_from_cam.inverse()
        trajectories[shot['timestamp'], RGBD_SENSOR_ID] = pose_cam_from_world

    # sensors
    """
    Read info.txt
    """
    info_filepath = path.join(d12scenes_path, 'info.txt')
    if not path.isfile(info_filepath):
        raise FileNotFoundError(f'info file is missing: {info_filepath}.')

    with open(info_filepath, 'rt') as file:
        info_dict = {}
        for line in file.readlines():
            line_splits = line.rstrip().split(' = ')
            info_dict[line_splits[0]] = line_splits[1]

    sensors = kapture.Sensors()
    camera_type = kapture.CameraType.PINHOLE
    assert 'm_calibrationColorIntrinsic' in info_dict
    assert 'm_colorWidth' in info_dict
    assert 'm_colorHeight' in info_dict
    rgb_intrinsics = [float(v) for v in info_dict['m_calibrationColorIntrinsic'].split(' ')]
    # w, h, fx, fy, cx, cy
    rgb_camera_params = [int(info_dict['m_colorWidth']), int(info_dict['m_colorHeight']),
                         rgb_intrinsics[0], rgb_intrinsics[5], rgb_intrinsics[2], rgb_intrinsics[6]]
    sensors[RGB_SENSOR_ID] = kapture.Camera(
        name=RGB_SENSOR_ID,
        camera_type=camera_type,
        camera_params=rgb_camera_params
    )

    assert 'm_calibrationDepthIntrinsic' in info_dict
    assert 'm_depthWidth' in info_dict
    assert 'm_depthHeight' in info_dict
    depth_intrinsics = [float(v) for v in info_dict['m_calibrationDepthIntrinsic'].split(' ')]
    # w, h, fx, fy, cx, cy
    depth_camera_params = [int(info_dict['m_depthWidth']), int(info_dict['m_depthHeight']),
                           depth_intrinsics[0], depth_intrinsics[5], depth_intrinsics[2], depth_intrinsics[6]]
    sensors[DEPTH_SENSOR_ID] = kapture.Camera(
        name=DEPTH_SENSOR_ID,
        camera_type=camera_type,
        camera_params=depth_camera_params,
        sensor_type='depth'
    )

    sensors[REG_DEPTH_SENSOR_ID] = kapture.Camera(
        name=REG_DEPTH_SENSOR_ID,
        camera_type=camera_type,
        camera_params=rgb_camera_params,
        sensor_type='depth'
    )

    # bind camera and depth sensor into a rig
    logger.info('building rig with camera and depth sensor ...')
    rigs = kapture.Rigs()
    rigs[RGBD_SENSOR_ID, RGB_SENSOR_ID] = kapture.PoseTransform()
    rigs[RGBD_SENSOR_ID, DEPTH_SENSOR_ID] = kapture.PoseTransform()
    rigs[RGBD_SENSOR_ID, REG_DEPTH_SENSOR_ID] = kapture.PoseTransform()

    # import (copy) image files.
    logger.info('copying image files ...')
    image_filenames = [f for _, _, f in kapture.flatten(snapshots)]
    import_record_data_from_dir_auto(d12images_path, kapture_dir_path, image_filenames, images_import_method)

    # import (copy) depth map files.
    logger.info('converting depth files ...')
    depth_map_filenames = kapture.io.records.records_to_filepaths(depth_maps, kapture_dir_path)
    hide_progress = logger.getEffectiveLevel() > logging.INFO
    for depth_map_filename, depth_map_filepath_kapture in tqdm(depth_map_filenames.items(), disable=hide_progress):
        if '.reg' in depth_map_filename:
            continue
        depth_map_filepath_12scenes = path.join(d12images_path, depth_map_filename + '.png')
        depth_map = np.array(Image.open(depth_map_filepath_12scenes))
        # depth maps is in mm in 12scenes, convert it to meters
        depth_map = depth_map.astype(np.float32) * 1.0e-3
        kapture.io.records.records_depth_to_file(depth_map_filepath_kapture, depth_map)
        # register depth to rgb
        reg_depth_map = register_depth(get_K(camera_type, depth_camera_params), get_K(camera_type, rgb_camera_params),
                                       np.eye(4), depth_map, rgb_camera_params[0], rgb_camera_params[1])
        kapture.io.records.records_depth_to_file(depth_map_filepath_kapture + '.reg', reg_depth_map)

    # pack into kapture format
    imported_kapture = kapture.Kapture(
        records_camera=snapshots,
        records_depth=depth_maps,
        rigs=rigs,
        trajectories=trajectories,
        sensors=sensors)

    logger.info('writing imported data ...')
    kapture_to_dir(kapture_dir_path, imported_kapture)


def import_12scenes_command_line() -> None:
    """
    Imports RGB-D Dataset 12-Scenes dataset and save them as kapture using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(
        description='Imports RGB-D Dataset 12-Scenes files to the kapture format.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-i', '--input', required=True,
                        help='input path Dataset 12-Scenes sequence root path')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.link_absolute,
                        help=f'How to import images [link_absolute], '
                             f'choose among: {", ".join(a.name for a in TransferAction)}')
    parser.add_argument('-o', '--output', required=True, help='output directory.')
    parser.add_argument('-p', '--partition', default=None, choices=['mapping', 'query'],
                        help='limit to mapping or query sequences only (using authors split files).')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    import_12scenes(d12scenes_path=args.input,
                    kapture_dir_path=args.output,
                    force_overwrite_existing=args.force,
                    images_import_method=args.image_transfer,
                    partition=args.partition)


if __name__ == '__main__':
    import_12scenes_command_line()



# Copyright 2021-present NAVER Corp. Under BSD 3-clause license

"""
Computation helper operations.
"""


def num_digits(n) -> int:
    """
    Compute the number of digits of a number

    :return: number of digits
    """
    if not isinstance(n, int):
        raise TypeError(f'Invalid n: not an integer {n}')
    count = 1
    while (int(n / 10)) != 0:
        count += 1
        n = int(n / 10)
    return count



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

from .PoseTransform import PoseTransform
from typing import Dict, Union, Tuple, List


class Rigs(Dict[str, Dict[str, PoseTransform]]):
    """
    brief: Rigs is a set of rig. Each rig is uniquely defined by a sensor_id.
        rigs[rig_id][sensor_id] = <PoseTransform>
        or
        rigs[rig_id, sensor_id] = <PoseTransform>

        with <PoseTransform> transforming points from rig to camera.
    """

    def __setitem__(self,
                    key: Union[str, Tuple[str, str]],
                    value: Union[Dict[str, PoseTransform], PoseTransform]):
        # enforce type checking
        if isinstance(key, tuple):
            # rigs[rig_id, sensor_id] = <PoseTransform>
            assert len(key) == 2
            if any(not isinstance(x, str) for x in key):
                raise TypeError('expect str type as key pair')
            rig_device_id, sensor_id = key[0], key[1]
            if not isinstance(value, PoseTransform):
                raise TypeError('expect PoseTransform type as value')
            self.setdefault(rig_device_id, {})[sensor_id] = value

        elif isinstance(key, str):
            # key is rig_id
            rig_id = key
            # rigs[rig_id] = {}
            # check type of values?
            if not isinstance(value, dict):
                raise TypeError('invalid value for rig id.')
            if not all(isinstance(k, str) for k in value.keys()):
                raise TypeError('invalid device_id')
            if not all(isinstance(v, PoseTransform) for v in value.values()):
                raise TypeError('invalid Pose')
            super(Rigs, self).__setitem__(rig_id, value)
        else:
            raise TypeError('invalid key type for Rigs')

    def __getitem__(self, key: Union[str, Tuple[str, str]]) -> Union[Dict[str, PoseTransform], PoseTransform]:
        if isinstance(key, str):
            # rigs[rig_id] = <Rig>
            return super(Rigs, self).__getitem__(key)
        elif isinstance(key, tuple):
            assert len(key) == 2
            rig_id, sensor_id = key
            if not isinstance(rig_id, str):
                raise TypeError('invalid rig_id')
            if not isinstance(sensor_id, str):
                raise TypeError('invalid sensor_id')
            return super(Rigs, self).__getitem__(rig_id)[sensor_id]
        else:
            raise TypeError('key must be either str or Tuple[str, str]')

    def key_pairs(self) -> List[Tuple[str, str]]:
        """
        Returns the list of (rig_id, device_id) contained in trajectories.
        Those pairs can be used to access a single trajectory pose.

        :return: list of (rig_id, device_id)
        """
        return [
            (rig_id, sensor_id)
            for rig_id, sensors in self.items()
            for sensor_id in sensors.keys()
        ]

    def __repr__(self) -> str:
        # [rig_id, sensor_id] = qw, qx, qy, qz, tx, ty, tz
        poses = [f'[{rig_id:5}, {sensor_id:5}] = {pose}'
                 for rig_id, rig in self.items()
                 for sensor_id, pose in rig.items()]
        return '\n'.join(poses)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Openmvg constants values useful for conversion from and to.
"""

from kapture.utils.Collections import AutoEnum
from enum import auto
from dataclasses import dataclass

OPENMVG_DEFAULT_JSON_FILE_NAME = 'sfm_data.json'
OPENMVG_DEFAULT_REGIONS_FILE_NAME = 'image_describer.json'
OPENMVG_SFM_DATA_VERSION_NUMBER = "0.3"


# XML names
class JSON_KEY:
    SFM_DATA_VERSION = "sfm_data_version"
    ROOT_PATH = "root_path"
    VIEWS = "views"
    VIEW_PRIORS = "view_priors"
    KEY = "key"
    VALUE = "value"
    POLYMORPHIC_ID = "polymorphic_id"
    PTR_WRAPPER = "ptr_wrapper"
    ID = "id"
    DATA = "data"
    LOCAL_PATH = "local_path"
    FILENAME = "filename"
    WIDTH = "width"
    HEIGHT = "height"
    ID_VIEW = "id_view"
    ID_INTRINSIC = "id_intrinsic"
    ID_POSE = "id_pose"
    INTRINSICS = "intrinsics"
    POLYMORPHIC_NAME = "polymorphic_name"
    VALUE0 = "value0"
    FOCAL_LENGTH = "focal_length"
    PRINCIPAL_POINT = "principal_point"
    DISTO_K1 = "disto_k1"
    DISTO_K3 = "disto_k3"
    DISTO_T2 = "disto_t2"
    FISHEYE = "fisheye"
    EXTRINSICS = "extrinsics"
    USE_POSE_ROTATION_PRIOR = "use_pose_rotation_prior"
    ROTATION_WEIGHT = "rotation_weight"
    ROTATION = "rotation"
    USE_POSE_CENTER_PRIOR = "use_pose_center_prior"
    CENTER_WEIGHT = "center_weight"
    CENTER = "center"
    STRUCTURE = "structure"
    CONTROL_POINTS = "control_points"


# Camera models
class CameraModel(AutoEnum):
    """
    Enumeration that contains all the openmvg camera model
    """
    pinhole = auto()
    pinhole_radial_k1 = auto()
    pinhole_radial_k3 = auto()
    pinhole_brown_t2 = auto()
    fisheye = auto()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

import requests
import os
import os.path as path
from tqdm import tqdm
from typing import Optional
import logging

logger = logging.getLogger('downloader')


def get_remote_file_size(url: str):
    """
    return the total file size on the remote url.

    :param url: input full url of the file.
    :return : int of file size in bytes, or None if unknown
    """
    # hack: use requests.get(range0-10) instead of requests.header. Header does not follow redirections in some cases.
    headers = {"Range": "bytes=0-10"}  # only retrieve the first bytes,
    response = requests.get(url, allow_redirects=True, headers=headers)
    # content length is not relevant (always 11, since that is what was requested)
    # lets look to Content-Range, that is formatted as 0-11/total
    if 'Content-Range' not in response.headers or '/' not in response.headers['Content-Range']:
        return None
    nb_bytes_total = int(response.headers['Content-Range'].split('/')[1])
    return nb_bytes_total


def download_file_resume(url: str,
                         filepath: str,
                         resume_byte_pos: Optional[int] = None):
    """
    resume (or start if no pos given) the dataset download.

    :param url: input full url of the file to be downloaded.
    :param filepath: input full path where to save the file.
    :param resume_byte_pos: input position in bytes where to resume the Download
    """

    file_size_online = get_remote_file_size(url)
    # Append information to resume download at specific byte position to header
    resume_header = ({'Range': f'bytes={resume_byte_pos}-'}
                     if resume_byte_pos else None)
    # Establish connection
    response = requests.get(url, stream=True, headers=resume_header, allow_redirects=True)
    # Set configuration
    block_size = 1024  # 1Ko
    initial_pos = resume_byte_pos if resume_byte_pos else 0
    mode = 'ab' if resume_byte_pos else 'wb'
    hide_progress = logger.getEffectiveLevel() > logging.INFO
    with open(filepath, mode) as f:
        with tqdm(total=file_size_online, unit='B',
                  unit_scale=True, unit_divisor=block_size,
                  desc=filepath, initial=initial_pos,
                  ascii=True, miniters=1,
                  disable=hide_progress) as pbar:
            for chunk in response.iter_content(32 * block_size):
                f.write(chunk)
                pbar.update(len(chunk))


def download_file(url, filepath):
    """
     Starts or resumes the download if already started.

    :param url: input full url of the file to be downloaded.
    :param filepath: input full path where to save the file.
     """
    resume_position = None
    if path.isfile(filepath):
        logger.debug('file is already (partially) there.')
        # file already there (at least partially)
        file_size_online = get_remote_file_size(url)
        file_size_local = int(path.getsize(filepath))
        if file_size_online is None:
            raise ValueError('Unable to retrieve file size on remote url.')

        if file_size_online == file_size_local:
            logger.debug(f'file {filepath} already downloaded.')
            return

        logger.debug(f'resume download from {file_size_local / file_size_online * 100.:4.1f}%')
        resume_position = file_size_local

    logger.debug(f'start downloading "{filepath}"\n\tfrom: "{url}".')
    os.makedirs(path.dirname(filepath), exist_ok=True)
    download_file_resume(url, filepath, resume_position)



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to print statistics about kapture data.
"""

import argparse
import logging
import os
import os.path as path
import re
from typing import Optional, Tuple
from datetime import datetime
import time

import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.computation as computation
import kapture.utils.logging
import kapture.io.csv

logger = logging.getLogger('kapture_data_date')

KAPTURE_FILE_PARSING_RE = re.compile('.*(records|trajectories).*\\.txt')


def format_timestamp_range(first_timestamp: int, last_timestamp: int) -> str:
    """
    Format a time range with unit.

    :param first_timestamp: first timestamp
    :param last_timestamp: last timestamp
    :return string: formatted time range
    """
    timestamp_len = 1
    try:
        timestamp_len = computation.num_digits(first_timestamp)
        delta_len = 10 - timestamp_len  # datetime expects epoch timestamp in seconds
        factor = 10**delta_len
        dts = [datetime.utcfromtimestamp(first_timestamp * factor),
               datetime.utcfromtimestamp(last_timestamp * factor)]
        timestamp_format_d = {
            'date': '%Y/%m/%d',
            'time': '%H:%M:%S.%f'
        }
        timestamp_parts = [
            {
                pname: dt.strftime(pformat)
                for pname, pformat in timestamp_format_d.items()
            }
            for dt in dts
        ]
        timestamp_str = timestamp_parts[0]['date'] + ' ' + timestamp_parts[0]['time']
        timestamp_str += ' -> '
        if timestamp_parts[0]['date'] != timestamp_parts[1]['date']:
            timestamp_str = timestamp_parts[0]['date'] + ' '
        timestamp_str += timestamp_parts[1]['time'] + ' GMT'
        return timestamp_str
    except ValueError as _:  # noqa: F841
        return f'{first_timestamp} -> {last_timestamp}' \
               + f' ** FAIL to parse as posix timestamp of {timestamp_len} digits'


def print_info(kapture_path: str, kapture_filename: str) -> None:
    """
    Prints some info on given file that should contains records with timestamps

    :param kapture_path: kapture top directory
    :param kapture_filename: full path of a valid kapture file
    """
    # Check version to test if this is a real kapture file
    kapture_file_path = path.join(kapture_path, kapture_filename)
    version = kapture.io.csv.get_version_from_csv_file(kapture_file_path)
    if version is None:
        logger.debug(f'{kapture_filename} not a kapture file')
    else:
        # Read the file
        last_data_line = ''
        nb_lines = 0
        with open(kapture_file_path) as f:
            # Count lines
            for _ in f:
                nb_lines += 1
            # Reset to read some important lines
            f.seek(0, os.SEEK_SET)
            # Skip header
            f.readline()
            f.readline()
            first_data_line = f.readline()
            if first_data_line:
                last_data_line = kapture.io.csv.get_last_line(f)
        first_timestamp = 0
        last_timestamp = 0
        # data line are comma separated lines with the timestamp as first value
        try:
            if first_data_line:
                first_timestamp = int(first_data_line.split(',')[0])
            if last_data_line:
                last_timestamp = int(last_data_line.split(',')[0])
        except ValueError:
            pass
        if first_timestamp > 0 and last_timestamp > 0:
            timestamp_range_str = format_timestamp_range(first_timestamp, last_timestamp)
            timestamp_len1 = computation.num_digits(first_timestamp)
            timestamp_len2 = computation.num_digits(last_timestamp)
            timestamp_len_str = f'{timestamp_len1}' if timestamp_len1 == timestamp_len2\
                else f'{timestamp_len1}-{timestamp_len2}'
            print(f'{kapture_filename:42s} timestamp {timestamp_len_str} digits from {timestamp_range_str}'
                  f' : {(nb_lines-2):12,d} records'.replace(',', ' '))


def do_print(kapture_path: str) -> None:
    """
    Print out kapture data:

    :param kapture_path: full path to kapture directory.
    """
    dirs_to_examine = set()
    for k_path in kapture.io.csv.CSV_FILENAMES.values():
        dirs_to_examine.add(path.dirname(k_path))
    # Look for files which might contain kapture data
    for k_dir in sorted(dirs_to_examine):
        kapture_dir = path.join(kapture_path, k_dir)
        if path.exists(kapture_dir):
            if path.isdir(kapture_dir):
                # print
                logger.debug(f'browsing directory {k_dir}')
                # Search for records or trajectory like files
                kapture_files = []
                for entry in os.listdir(kapture_dir):
                    if path.isfile(path.join(kapture_dir, entry)):
                        if KAPTURE_FILE_PARSING_RE.match(entry) is not None:
                            kapture_files.append(entry)
                for kapture_file in sorted(kapture_files):
                    print_info(kapture_path, path.join(k_dir, kapture_file))
            else:
                logger.fatal(f'{kapture_dir} is not a directory')
        else:
            logger.debug(f'{kapture_dir} does not exist')


def print_command_line() -> None:
    """
    Do the print using the parameters given on the command line.
    """

    parser = argparse.ArgumentParser(description='Print statistics about kapture records files with timestamps.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-k', '--kapture', required=True,
                        help='path to kapture data root directory.')
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    do_print(path.abspath(args.kapture))


if __name__ == '__main__':
    print_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Merge kapture objects for reconstructions.
"""

from kapture.io.binary import array_to_file
from kapture.io.tar import TarCollection
import numpy as np
import os
import shutil
from typing import Dict, List, Union, Optional, Tuple, Type

import kapture
import kapture.io.features
from kapture.utils.logging import getLogger


def _merge_image_features(feature_class_type: Type[Union[kapture.Keypoints,
                                                         kapture.Descriptors,
                                                         kapture.GlobalFeatures]],
                          feature_type: str,
                          features_list: Union[List[Optional[kapture.Keypoints]],
                                               List[Optional[kapture.Descriptors]],
                                               List[Optional[kapture.GlobalFeatures]]],
                          features_paths: List[str],
                          output_path: str,
                          tar_handlers: List[TarCollection]
                          ) -> Union[kapture.Keypoints, kapture.Descriptors, kapture.GlobalFeatures]:
    """
    Merge several features_list (keypoints, descriptors or global features_list) (of same type) in one.

    :param feature_class_type: the type of features_list
    :param feature_type: the type (name of the folder) of the features
    :param features_list: the list of values
    :param features_paths: the paths
    :param output_path: root path of the features to construct
    :param tar_handlers: collection of preloaded tar archives
    :return: merged features object of the corresponding type
    """
    assert len(features_list) > 0
    assert len(features_paths) == len(features_list)

    # find no none value
    val = [(i, d) for i, d in enumerate(features_list) if d is not None]
    assert len(val) > 0

    merged_features = val[0][1]
    for j, (i, features) in enumerate(val):
        assert isinstance(features, feature_class_type)
        assert features.type_name == merged_features.type_name
        assert features.dtype == merged_features.dtype
        assert features.dsize == merged_features.dsize
        if feature_class_type == kapture.Descriptors or feature_class_type == kapture.GlobalFeatures:
            assert not isinstance(features, kapture.Keypoints)  # IDE type check help
            assert not isinstance(merged_features, kapture.Keypoints)  # IDE type check help
            assert features.metric_type == merged_features.metric_type
        if feature_class_type == kapture.Descriptors:
            assert isinstance(features, kapture.Descriptors)  # IDE type check help
            assert isinstance(merged_features, kapture.Descriptors)  # IDE type check help
            assert features.keypoints_type == merged_features.keypoints_type
        for name in features:
            if j > 0 and name in merged_features:
                getLogger().warning(f'{name} was found multiple times.')
            else:
                merged_features.add(name)
                if output_path:
                    # TODO: uses kapture.io.features_list.get_image_features_dirpath()
                    in_path = kapture.io.features.get_features_fullpath(feature_class_type,
                                                                        feature_type,
                                                                        features_paths[i],
                                                                        name,
                                                                        tar_handlers[i])
                    out_path = kapture.io.features.get_features_fullpath(feature_class_type,
                                                                         feature_type,
                                                                         output_path,
                                                                         name)
                    if in_path != out_path:
                        # skip actual copy if file does not actually move.
                        os.makedirs(os.path.dirname(out_path), exist_ok=True)
                        if isinstance(in_path, str):
                            shutil.copy(in_path, out_path)
                        else:
                            # in_path is a tuple [str, TarHandler]
                            # keypoints are not stored in a file, have to read them to be able to copy them
                            array = in_path[1].get_array_from_tar(in_path[0], features.dtype, features.dsize)
                            array_to_file(out_path, array)
    return merged_features


def _merge_image_features_collection(feature_class_type: Type[Union[kapture.Keypoints,
                                                                    kapture.Descriptors,
                                                                    kapture.GlobalFeatures]],
                                     features_list: Union[List[Optional[Dict[str, kapture.Keypoints]]],
                                                          List[Optional[Dict[str, kapture.Descriptors]]],
                                                          List[Optional[Dict[str, kapture.GlobalFeatures]]]],
                                     features_paths: List[str],
                                     output_path: str,
                                     tar_handlers: List[TarCollection]
                                     ) -> Union[Dict[str, kapture.Keypoints],
                                                Dict[str, kapture.Descriptors],
                                                Dict[str, kapture.GlobalFeatures]]:
    assert len(features_list) > 0
    assert len(features_paths) == len(features_list)
    # get the union
    features_types = set().union(*[features.keys() for features in features_list if features is not None])
    assert len(features_types) > 0
    out_collection = {}
    for features_type in features_types:
        image_features_list = [features[features_type] if features is not None and features_type in features else None
                               for features in features_list]
        image_features = _merge_image_features(feature_class_type, features_type,
                                               image_features_list,
                                               features_paths, output_path,
                                               tar_handlers)
        assert isinstance(image_features, feature_class_type)
        out_collection[features_type] = image_features
    return out_collection


def merge_keypoints(feature_type: str,
                    keypoints_list: List[Optional[kapture.Keypoints]],
                    keypoints_paths: List[str],
                    output_path: str,
                    tar_handlers: List[TarCollection]) -> kapture.Keypoints:
    """
    Merge several keypoints in one.

    :param keypoints_list: list of keypoints to merge
    :param keypoints_paths: keypoints files paths
    :param output_path: root path of the merged features files
    :param tar_handlers: collection of preloaded tar archives
    :return: merged keypoints
    """
    keypoints = _merge_image_features(kapture.Keypoints, feature_type, keypoints_list, keypoints_paths,
                                      output_path, tar_handlers)
    assert isinstance(keypoints, kapture.Keypoints)
    return keypoints


def merge_keypoints_collections(keypoints_collections_list: List[Optional[Dict[str, kapture.Keypoints]]],
                                keypoints_paths: List[str],
                                output_path: str,
                                tar_handlers: List[TarCollection]) -> Dict[str, kapture.Keypoints]:
    """
    Merge several keypoints collections in one.

    :param keypoints_collections_list: list of keypoints collections to merge
    :param keypoints_paths: keypoints files paths
    :param output_path: root path of the merged features files
    :param tar_handlers: collection of preloaded tar archives
    :return: merged keypoints collection
    """
    return _merge_image_features_collection(kapture.Keypoints, keypoints_collections_list,
                                            keypoints_paths, output_path, tar_handlers)


def merge_descriptors(feature_type: str,
                      descriptors_list: List[Optional[kapture.Descriptors]],
                      descriptors_paths: List[str], output_path: str,
                      tar_handlers: List[TarCollection]) -> kapture.Descriptors:
    """
    Merge several descriptors in one.

    :param descriptors_list: list of descriptors to merge
    :param descriptors_paths: descriptors files paths
    :param output_path: root path of the merged features files
    :param tar_handlers: collection of preloaded tar archives
    :return: merged descriptors
    """
    descriptors = _merge_image_features(kapture.Descriptors, feature_type,
                                        descriptors_list, descriptors_paths, output_path, tar_handlers)
    assert isinstance(descriptors, kapture.Descriptors)
    return descriptors


def merge_descriptors_collections(descriptors_collections_list: List[Optional[Dict[str, kapture.Descriptors]]],
                                  descriptors_paths: List[str],
                                  output_path: str,
                                  tar_handlers: List[TarCollection]) -> Dict[str, kapture.Descriptors]:
    """
    Merge several descriptors collections in one.

    :param descriptors_collections_list: list of descriptors collections to merge
    :param descriptors_paths: descriptors files paths
    :param output_path: root path of the merged features files
    :param tar_handlers: collection of preloaded tar archives
    :return: merged descriptors collections
    """
    return _merge_image_features_collection(kapture.Descriptors, descriptors_collections_list,
                                            descriptors_paths, output_path, tar_handlers)


def merge_global_features(global_features_list: List[Optional[kapture.GlobalFeatures]],
                          global_features_paths: List[str], output_path: str,
                          tar_handlers: List[TarCollection]) -> kapture.GlobalFeatures:
    """
    Merge several global features in one.

    :param global_features_list: list of global features to merge
    :param global_features_paths: global features files paths
    :param output_path: root path of the merged features files
    :param tar_handlers: collection of preloaded tar archives
    :return: merged global features
    """
    features = _merge_image_features(kapture.GlobalFeatures, global_features_list, global_features_paths,
                                     output_path, tar_handlers)
    assert isinstance(features, kapture.GlobalFeatures)
    return features


def merge_global_features_collections(global_features_collections_list: List[Optional[Dict[str,
                                                                                           kapture.GlobalFeatures]]],
                                      global_features_paths: List[str],
                                      output_path: str,
                                      tar_handlers: List[TarCollection]) -> Dict[str, kapture.GlobalFeatures]:
    """
    Merge several global features collections in one.

    :param global_features_collections_list: list of global features collections to merge
    :param global_features_paths: global features files paths
    :param output_path: root path of the merged features files
    :param tar_handlers: collection of preloaded tar archives
    :return: merged global features collection
    """
    return _merge_image_features_collection(kapture.GlobalFeatures, global_features_collections_list,
                                            global_features_paths, output_path, tar_handlers)


def merge_matches(keypoints_type: str,
                  matches_list: List[Optional[kapture.Matches]],
                  matches_paths: List[str],
                  output_path: str,
                  tar_handlers: List[TarCollection]) -> kapture.Matches:
    """
    Merge several matches lists in one.

    :param keypoints_type: type of keypoints, name of the keypoints subfolder
    :param matches_list: list of matches to merge
    :param matches_paths: matches files paths
    :param output_path: root path of the merged matches files
    :param tar_handlers: collection of preloaded tar archives
    :return: merged matches
    """
    assert len(matches_list) > 0
    assert len(matches_paths) == len(matches_list)

    merged_matches = kapture.Matches()
    for matches, matches_path, tar_handler in zip(matches_list, matches_paths, tar_handlers):
        if matches is None:
            continue
        for pair in matches:
            if pair in merged_matches:
                getLogger().warning(f'{pair} was found multiple times.')
            else:
                merged_matches.add(pair[0], pair[1])
                if output_path:
                    in_path = kapture.io.features.get_matches_fullpath(pair, keypoints_type, matches_path, tar_handler)
                    out_path = kapture.io.features.get_matches_fullpath(pair, keypoints_type, output_path)
                    if in_path != out_path:
                        # skip actual copy if file does not actually move.
                        os.makedirs(os.path.dirname(out_path), exist_ok=True)
                        if isinstance(in_path, str):
                            shutil.copy(in_path, out_path)
                        else:
                            # in_path is a tuple [str, TarHandler]
                            # keypoints are not stored in a file, have to read them to be able to copy them
                            array = kapture.io.features.image_matches_from_file(in_path)
                            kapture.io.features.image_matches_to_file(out_path, array)
    return merged_matches


def merge_matches_collections(matches_list: List[Optional[Dict[str,  kapture.Matches]]],
                              matches_paths: List[str],
                              output_path: str,
                              tar_handlers: List[TarCollection]) -> Dict[str,  kapture.Matches]:
    """
    Merge several matches collections in one.

    :param matches_list: list of matches collections to merge
    :param matches_paths: matches files paths
    :param output_path: root path of the merged matches files
    :param tar_handlers: collection of preloaded tar archives
    :return: merged matches collection
    """
    assert len(matches_list) > 0
    assert len(matches_paths) == len(matches_list)

    # get the union
    keypoints_types = set().union(*[matches.keys() for matches in matches_list if matches is not None])
    assert len(keypoints_types) > 0
    out_collection = {}
    for keypoints_type in keypoints_types:
        kmatches_list = [matches[keypoints_type] if matches is not None and keypoints_type in matches else None
                         for matches in matches_list]
        merged_matches = merge_matches(keypoints_type,
                                       kmatches_list,
                                       matches_paths,
                                       output_path,
                                       tar_handlers)
        assert isinstance(merged_matches, kapture.Matches)
        out_collection[keypoints_type] = merged_matches
    return out_collection


def merge_points3d_and_observations(pts3d_obs: List[Tuple[Optional[kapture.Points3d], Optional[kapture.Observations]]]
                                    ) -> Tuple[kapture.Points3d, kapture.Observations]:
    """
    Merge a list of points3d with their observations.

    :param pts3d_obs: list of points3d with observations to merge
    :return: merged points3d associated to observations
    """
    assert len(pts3d_obs) > 0
    merged_points3d = kapture.Points3d()
    merged_observations = kapture.Observations()
    point3d_offset = 0
    for points3d, observations in pts3d_obs:
        if points3d is None:
            continue
        merged_points3d = kapture.Points3d(np.vstack([merged_points3d, points3d]))
        if observations is not None:
            for point3d_idx, keypoints_type, (image_path, keypoint_idx) in kapture.flatten(observations):
                merged_observations.add(point3d_idx + point3d_offset, keypoints_type, image_path, keypoint_idx)
        point3d_offset += merged_points3d.shape[0]
    return merged_points3d, merged_observations


def merge_points3d(points3d_list: List[Optional[kapture.Points3d]]) -> kapture.Points3d:
    """
    Merge several points3d lists in one.

    :param points3d_list: list of points3d to merge
    :return: merged points3d
    """
    assert len(points3d_list) > 0
    merged_points3d = kapture.Points3d()
    for points3d in points3d_list:
        if points3d is None:
            continue
        merged_points3d = kapture.Points3d(np.vstack([merged_points3d, points3d]))
    return merged_points3d



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

import os
import os.path as path
from random import randint
from typing import Dict
import numpy as np

import kapture
from .csv import kapture_linesep
from .features import image_keypoints_from_file

PLY_HEADER_TEMPLATE = kapture_linesep.join([
    'ply',
    'format ascii 1.0',
    'element vertex {nb_vertex}',
    'property double x',
    'property double y',
    'property double z',
    'property uchar red',
    'property uchar green',
    'property uchar blue',
    'element edge {nb_edges}',
    'property int vertex1',
    'property int vertex2',
    'end_header'])

########################################################################################################################
BLACK = 3 * [0]
WHITE = 3 * [255]
GREY = 3 * [127]
RED = [255, 0, 0]
GREEN = [0, 255, 0]
BLUE = [0, 0, 255]
AXIS_COLORS = [GREY, RED, GREEN, BLUE]


########################################################################################################################
def get_axis_in_world(
        pose_device_from_world: kapture.PoseTransform, length: float = 1.0) -> kapture.Points3d:
    """
    Returns a quadruplet of points (0,x,y,z) representing the axis of the device into the world.

    :param pose_device_from_world: assume the transformation is device from world.
    :param length: distance between each axis point and center.
    :return: 4 points (center, x, y, z) arranged by cols
    """
    assert (isinstance(pose_device_from_world, kapture.PoseTransform))
    sensor_axis = np.array([
        [0, 0, 0],  # 0
        [length, 0, 0],  # X
        [0, length, 0],  # Y
        [0, 0, length],  # Z
    ], dtype=np.float)
    return pose_device_from_world.inverse().transform_points(sensor_axis)


def header_to_ply_stream(stream, nb_vertex: int = 0, nb_edges: int = 0) -> None:
    """
    Writes PLY header to a stream.

    :param stream: an open stream to write to
    :param nb_vertex: number of vertex
    :param nb_edges: number of edges
    """
    stream.write(PLY_HEADER_TEMPLATE.format(nb_vertex=nb_vertex,
                                            nb_edges=nb_edges))
    stream.write(kapture_linesep)


def rig_to_ply_stream(stream, rig: Dict[str, kapture.PoseTransform], axis_length: float = 1.) -> None:
    """
    Writes the rig to a stream.

    :param stream: an open stream to write to
    :param rig: rig to write
    :param axis_length: length of the axis
    """
    device_list = [(cam_id, pose_tr)
                   for cam_id, pose_tr in rig.items()]

    # add the rig center in devices
    device_list = [(-1, kapture.PoseTransform())] + device_list

    # create 4 points per device: 1 for center, 3 for axis
    points_colored_list = []
    edges_list = []
    for cam_id, pose_tr in device_list:
        axis = get_axis_in_world(pose_tr, axis_length)
        edges_list += [(0, len(points_colored_list))]
        points_colored_list += [p + AXIS_COLORS[i] for i, p in enumerate(axis.tolist())]

    # write points into ply
    header_to_ply_stream(stream,
                         nb_vertex=len(points_colored_list),
                         nb_edges=len(edges_list))
    for p3d in points_colored_list:
        line = ['{:<25}'.format(i) for i in p3d[0:3]]
        line += ['{:03}'.format(int(i)) for i in p3d[3:6]]
        stream.write(' '.join(line) + kapture_linesep)
    for e in edges_list:
        line = ['{:2}'.format(i) for i in e]
        stream.write(' '.join(line) + kapture_linesep)


def rig_to_ply(filepath: str, rig: Dict[str, kapture.PoseTransform], axis_length: float = 1.) -> None:
    """
    Writes the rig to a file.

    :param filepath: file path to write to
    :param rig: rig to write
    :param axis_length: length of the axis
    """
    with open(filepath, 'w') as f:
        rig_to_ply_stream(f, rig, axis_length)


########################################################################################################################
def trajectories_to_ply_stream(stream, trajectories: kapture.Trajectories, axis_length: float = 1.) -> None:
    """
    Writes the trajectories to a stream.
     trajectories[ts][device_id] = [pose]

    :param stream: an open stream to write to
    :param trajectories: trajectories to write
    :param axis_length: length of the axis
    """
    pose_list = (pose_tr
                 for _, _, pose_tr in kapture.flatten(trajectories, is_sorted=True)
                 if not np.any(np.isnan(pose_tr.t)))  # filter out if no position

    # create 4 points per pose: 1 for center, 3 for axis
    points_colored_list = []
    for pose_tr in pose_list:
        axis = get_axis_in_world(pose_tr, axis_length)
        points_colored_list += [p + AXIS_COLORS[i] for i, p in enumerate(axis.tolist())]

    # write points into ply
    header_to_ply_stream(stream, nb_vertex=len(points_colored_list))
    for p3d in points_colored_list:
        line = ['{:<25}'.format(i) for i in p3d[0:3]]
        line += ['{:<4}'.format(i) for i in p3d[3:6]]
        stream.write(' '.join(line) + kapture_linesep)


def trajectories_to_ply(
        filepath: str,
        trajectories: kapture.Trajectories,
        axis_length: float = 1.
):
    """
    Writes trajectory to PLY format (for visualization).
    Each pose in trajectory leads to a ply dot. 3 additional points are added in X (red), Y (Green) and Z (blue)
    direction around each pose.

    :param filepath: input ply file path.
    :param trajectories: input trajectory
    :param axis_length: length of axis representing the orientation of each pose in trajectory.
    :return:
    """
    os.makedirs(path.dirname(filepath), exist_ok=True)
    with open(filepath, 'w') as fout:
        trajectories_to_ply_stream(fout, trajectories, axis_length)


def points3d_to_stream(stream, points3d: kapture.Points3d) -> None:
    """
    Writes the 3D points to a stream.

    :param stream: an open stream to write to
    :param points3d: 3d points to write
    """
    header_to_ply_stream(stream, nb_vertex=len(points3d))
    for p3d in points3d:
        line = ['{:20}'.format(i) for i in p3d[0:3]] + ['{:03d}'.format(int(i)) for i in p3d[3:6]]
        stream.write('  '.join(line) + kapture_linesep)


def points3d_to_ply(filepath: str, points3d: kapture.Points3d) -> None:
    """
    Writes 3D points into ply file.

    :param filepath: ply file path.
    :param points3d: 3D points.
    """
    os.makedirs(path.dirname(filepath), exist_ok=True)
    with open(filepath, 'w') as f:
        points3d_to_stream(f, points3d)


def image_keypoints_to_stream(stream, image_keypoints: np.array) -> None:
    """
    Plots image keypoints onto a 2D plane. Use random colors.

    :param stream: an open stream to write to
    :param image_keypoints: the image keypoints to write
    """
    header_to_ply_stream(stream, nb_vertex=len(image_keypoints))
    for kpt in image_keypoints:
        coords = kpt[0:2]
        line = ['{:20}'.format(i) for i in coords] + ['0.0'] + ['{:03d}'.format(randint(0, 255)) for _ in range(3)]
        stream.write('  '.join(line) + kapture_linesep)


def image_keypoints_to_ply(ply_filepath: str, image_keypoints_filepath: str, keypoint_dtype, keypoint_dsize) -> None:
    """
    Plots image keypoints onto a 2D plane.

    :param ply_filepath: path to the ply file to write
    :param image_keypoints_filepath: path to the image keypoints file to read
    :param keypoint_dtype: keypoint data type
    :param keypoint_dsize: keypoint data size
    """
    os.makedirs(path.dirname(ply_filepath), exist_ok=True)
    image_keypoints = image_keypoints_from_file(image_keypoints_filepath, keypoint_dtype, keypoint_dsize)
    with open(ply_filepath, 'w') as f:
        image_keypoints_to_stream(f, image_keypoints)



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to import a bundler model into a kapture.
A bundler model file is defined here:

    http://www.cs.cornell.edu/~snavely/bundler/bundler-v0.4-manual.html#S6

    # Bundle file v0.3
    <num_cameras> <num_points>   [two integers]
    <camera1>
    <camera2>
       ...
    <cameraN>
    <point1>
    <point2>
       ...
    <pointM>

    Each camera entry <cameraI> contains the estimated camera intrinsics and extrinsics, and has the form:


        <f> <k1> <k2>   [the focal length, followed by two radial distortion coeffs]
        <R>             [a 3x3 matrix representing the camera rotation]
        <t>             [a 3-vector describing the camera translation]

    The cameras are specified in the order they appear in the list of images.

    Each point entry has the form:


        <position>      [a 3-vector describing the 3D position of the point]
        <color>         [a 3-vector describing the RGB color of the point]
        <view list>     [a list of views the point is visible in]


"""

import argparse
import logging
import os
import os.path as path
import numpy as np
import quaternion
from PIL import Image
# kapture
import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.csv import kapture_to_dir
import kapture.io.features
from kapture.io.records import TransferAction, import_record_data_from_dir_auto

logger = logging.getLogger('bundler')
MODEL = kapture.CameraType.RADIAL
LOCAL_FEATURE_TYPE = 'sift'


def import_bundler(bundler_path: str,
                   image_list_path: str,
                   image_dir_path: str,
                   kapture_dir_path: str,
                   ignore_trajectories: bool,
                   add_reconstruction: bool,
                   force_overwrite_existing: bool = False,
                   images_import_method: TransferAction = TransferAction.skip) -> None:
    """
    Imports bundler data and save them as kapture.

    :param bundler_path: path to the bundler model file
    :param image_list_path: path to the file containing the list of image names
    :param image_dir_path: input path to bundler image directory.
    :param kapture_dir_path: path to kapture top directory
    :param ignore_trajectories: if True, will not import the trajectories
    :param add_reconstruction: if True, will create 3D points and observations
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    :param images_import_method: choose how to import actual image files.
    """
    os.makedirs(kapture_dir_path, exist_ok=True)
    delete_existing_kapture_files(kapture_dir_path, force_erase=force_overwrite_existing)

    logger.info('loading all content...')
    # if there is a filter list, parse it
    with open(image_list_path) as file:
        file_content = file.readlines()
    # remove end line char and empty lines
    image_list = [line.rstrip() for line in file_content if line != '\n']

    with open(bundler_path) as file:
        bundler_content = file.readlines()
    # remove end line char and empty lines
    bundler_content = [line.rstrip() for line in bundler_content if line != '\n']
    assert bundler_content[0] == "# Bundle file v0.3"
    # <num_cameras> <num_points>
    line_1 = bundler_content[1].split()
    number_of_cameras = int(line_1[0])
    number_of_points = int(line_1[1])
    offset = 2
    number_of_lines_per_camera = 5  # 1 camera + 3 rotation + 1 translation

    cameras = kapture.Sensors()
    images = kapture.RecordsCamera()
    trajectories = kapture.Trajectories() if not ignore_trajectories else None
    points3d = [] if add_reconstruction else None
    keypoints = None
    observations = kapture.Observations() if add_reconstruction else None
    image_mapping = []  # bundler camera_id -> (name, width, height)
    for i in range(0, number_of_cameras):
        start_index = i * number_of_lines_per_camera + offset
        file_name = image_list[i]

        # process camera info
        line_camera = bundler_content[start_index].split()
        focal_length = float(line_camera[0])
        k1 = float(line_camera[1])
        k2 = float(line_camera[2])

        # lazy open
        with Image.open(path.join(image_dir_path, file_name)) as im:
            width, height = im.size

        image_mapping.append((file_name, width, height))
        camera = kapture.Camera(MODEL, [width, height, focal_length, width / 2, height / 2, k1, k2])
        camera_id = f'sensor{i}'
        cameras[camera_id] = camera

        # process extrinsics
        rotation_matrix = [[float(v) for v in line.split()]
                           for line in bundler_content[start_index + 1:start_index + 4]]

        quaternion_wxyz = quaternion.from_rotation_matrix(rotation_matrix)
        translation = np.array([float(v) for v in bundler_content[start_index + 4].split()])
        pose = kapture.PoseTransform(quaternion_wxyz, translation)

        # The Bundler model uses a coordinate system that differs from the *computer vision camera
        #  coordinate system*. More specifically, they use the camera coordinate system typically used
        #  in *computer graphics*. In this camera coordinate system, the camera is looking down the
        #  `-z`-axis, with the `x`-axis pointing to the right and the `y`-axis pointing upwards.
        # rotation Pi around the x axis to get the *computer vision camera
        #  coordinate system*
        rotation_around_x = quaternion.quaternion(0.0, 1.0, 0.0, 0.0)
        transformation = kapture.PoseTransform(rotation_around_x, np.array([0, 0, 0]))

        images[(i, camera_id)] = file_name
        if trajectories is not None:
            # transformation.inverse() is equal to transformation (rotation around -Pi or Pi around X is the same)
            trajectories[(i, camera_id)] = kapture.PoseTransform.compose([transformation, pose, transformation])

    if points3d is not None and number_of_points > 0:
        assert observations is not None
        offset += number_of_cameras * number_of_lines_per_camera
        number_of_lines_per_point = 3  # position color viewlist

        # (image_name, bundler_keypoint_id ) -> keypoint_id
        known_keypoints = {}
        local_keypoints = {}

        for i in range(0, number_of_points):
            start_index = i * number_of_lines_per_point + offset
            position = [float(v) for v in bundler_content[start_index].split()]
            # apply transformation
            position = [position[0], -position[1], -position[2]]
            color = [float(v) for v in bundler_content[start_index + 1].split()]

            # <view list>: length of the list + [<camera> <key> <x> <y>]
            # x, y origin is the center of the image
            view_list = bundler_content[start_index + 2].split()
            number_of_observations = int(view_list[0])

            for j in range(number_of_observations):
                camera_id = int(view_list[1 + 4 * j + 0])
                keypoint_id = int(view_list[1 + 4 * j + 1])
                x = float(view_list[1 + 4 * j + 2])
                y = float(view_list[1 + 4 * j + 3])

                file_name, width, height = image_mapping[camera_id]
                # put (0,0) in upper left corner
                x += (width / 2)
                y += (height / 2)

                # init local_keypoints if needed
                if file_name not in local_keypoints:
                    local_keypoints[file_name] = []
                # do not add the same keypoint twice
                if (file_name, keypoint_id) not in known_keypoints:
                    # in the kapture format, keypoint id is different. Note that it starts from 0
                    known_keypoints[(file_name, keypoint_id)] = len(local_keypoints[file_name])
                    local_keypoints[file_name].append([x, y])
                keypoint_idx = known_keypoints[(file_name, keypoint_id)]
                observations.add(i, LOCAL_FEATURE_TYPE, file_name, keypoint_idx)
            points3d.append(position + color)
        points3d = np.array(points3d)

        # finally, convert local_keypoints to np.ndarray and add them to the global keypoints variable
        sift_keypoints = kapture.Keypoints(LOCAL_FEATURE_TYPE, np.float32, 2)
        for image_filename, keypoints_array in local_keypoints.items():
            keypoints_np_array = np.array(keypoints_array, dtype=np.float32)
            keypoints_out_path = kapture.io.features.get_keypoints_fullpath(LOCAL_FEATURE_TYPE,
                                                                            kapture_dir_path,
                                                                            image_filename)
            kapture.io.features.image_keypoints_to_file(keypoints_out_path, keypoints_np_array)
            sift_keypoints.add(image_filename)
        keypoints = {LOCAL_FEATURE_TYPE: sift_keypoints}
    if points3d is not None:
        points3d = kapture.Points3d(points3d)

    # import (copy) image files.
    logger.info('import image files ...')
    filename_list = [f for _, _, f in kapture.flatten(images)]
    import_record_data_from_dir_auto(image_dir_path, kapture_dir_path, filename_list, images_import_method)

    # pack into kapture format
    imported_kapture = kapture.Kapture(sensors=cameras, records_camera=images, trajectories=trajectories,
                                       points3d=points3d, keypoints=keypoints, observations=observations)
    logger.info('writing imported data...')
    kapture_to_dir(kapture_dir_path, imported_kapture)


def bundler_command_line() -> None:
    """
    Do the bundler import to kapture using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(
        description='import bundler file to the kapture format.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-i', '--input', required=True,
                        help='input path to .out bundler file')
    parser.add_argument('-l', '--image-list', required=True,
                        help='path to the list of images in the bundler order (.list.txt)')
    parser.add_argument('-im', '--image-path', required=True, help='path to images')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.link_absolute,
                        help=f'How to import images [link_absolute], '
                             f'choose among: {", ".join(a.name for a in TransferAction)}')
    parser.add_argument('-o', '--output', required=True, help='output directory.')
    parser.add_argument('--ignore-trajectories', action='store_true', default=False,
                        help='Do not export extrinsics.')
    parser.add_argument('--add-reconstruction', action='store_true', default=False,
                        help='add the 3d points/keypoints/observations to the output')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    import_bundler(bundler_path=args.input,
                   image_list_path=args.image_list,
                   image_dir_path=args.image_path,
                   kapture_dir_path=args.output,
                   ignore_trajectories=args.ignore_trajectories,
                   add_reconstruction=args.add_reconstruction,
                   force_overwrite_existing=args.force,
                   images_import_method=args.image_transfer)


if __name__ == '__main__':
    bundler_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
All base kapture objects, with their reading and writing functions on disk,
and the conversion from and to other formats.
"""

from .core import *  # noqa: F401 F403
# silence kapture logging to critical only, except if told otherwise
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.CRITICAL)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

from typing import Tuple, Dict, List, Set, Any, Iterable, Union


def flatten(element: Union[Any, Dict[Any, Any], List[Any], Set[Any]],
            is_sorted: bool = False) -> Iterable[Tuple[Any, ...]]:
    """
    flatten any nested dictionary into a list of tuples:
    {k1: {l1: v1, l2: v2}, k2: {l3: v3}} -> [(k1, l1, v1), (k1, l2, v2), (k2, l3, v3)]
    or
    {k1: [l1, l2], k2: [v1, v2]} -> [(k1, l1), (k1, l2), (k2, v1), (k2, v2)]
    """
    def sort_func(a): return sorted(a) if is_sorted else a
    if isinstance(element, dict):
        for key, value in sort_func(element.items()):
            for tup in flatten(value, is_sorted):
                yield (key,) + tup
    elif isinstance(element, list) or isinstance(element, set):
        for value in sort_func(element):
            for tup in flatten(value, is_sorted):
                yield tup
    else:
        yield element,



from kapture.utils.paths import path_secure
import kapture
import tarfile
import numpy as np
import io
from typing import Any, Dict, Iterable, List, Optional, Type, Union
import os.path as path


logger = kapture.logger


class TarHandler:
    def __init__(self, tarfile_path: str, mode: str = 'r'):
        assert mode in {'r', 'a'}
        self.mode = mode
        self.fid = tarfile.TarFile(tarfile_path, mode=self.mode)
        # list all files
        tarcontent = self.fid.getmembers()
        # if c.name is found multiple time, the value will correspond to its last occurence, so the most up to data one
        self.content = {path_secure(c.name): c for c in tarcontent}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def flush(self):
        self.fid.fileobj.flush()

    def close(self):
        self.fid.close()

    def add_array_to_tar(self, filepath: str, data_array: np.ndarray) -> None:
        assert self.mode == 'a'
        filepath = path_secure(filepath)
        info = tarfile.TarInfo(filepath)
        data = data_array.tobytes()
        info.size = len(data)
        self.fid.addfile(tarinfo=info, fileobj=io.BytesIO(data))
        self.flush()
        # self.content[filepath] = info
        self.content[filepath] = self.fid.getmember(filepath)

    def get_array_from_tar(self, filepath: str, dtype: Type, dsize: int) -> np.ndarray:
        assert self.mode == 'r'
        filepath = path_secure(filepath)
        if filepath not in self.content:
            print(self.content)
        info = self.content[filepath]
        data_array = np.frombuffer(self.fid.extractfile(info).read(), dtype=dtype)
        data_array = data_array.reshape((-1, dsize))
        return data_array


class TarCollection:

    def __init__(self,
                 keypoints: Optional[Dict[str, TarHandler]] = None,
                 descriptors: Optional[Dict[str, TarHandler]] = None,
                 global_features: Optional[Dict[str, TarHandler]] = None,
                 matches: Optional[Dict[str, TarHandler]] = None):
        super().__init__()
        if keypoints is None:
            self.keypoints = {}
        else:
            self.keypoints = keypoints

        if descriptors is None:
            self.descriptors = {}
        else:
            self.descriptors = descriptors

        if global_features is None:
            self.global_features = {}
        else:
            self.global_features = global_features

        if matches is None:
            self.matches = {}
        else:
            self.matches = matches

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        """
        closes all handlers
        """
        for _, handler in self.keypoints.items():
            handler.close()
        self.keypoints.clear()
        for _, handler in self.descriptors.items():
            handler.close()
        self.descriptors.clear()
        for _, handler in self.global_features.items():
            handler.close()
        self.global_features.clear()
        for _, handler in self.matches.items():
            handler.close()
        self.matches.clear()

    @property
    def keypoints(self) -> Dict[str, TarHandler]:
        """
        :return: the keypoints tar handlers
        """
        return self._keypoints

    @keypoints.setter
    def keypoints(self, keypoints: Dict[str, TarHandler]):
        if not isinstance(keypoints, dict):
            raise TypeError('dict of TarHandler expected')
        self._keypoints = keypoints

    @property
    def descriptors(self) -> Dict[str, TarHandler]:
        """
        :return: the descriptors tar handlers
        """
        return self._descriptors

    @descriptors.setter
    def descriptors(self, descriptors: Dict[str, TarHandler]):
        if not isinstance(descriptors, dict):
            raise TypeError('dict of TarHandler expected')
        self._descriptors = descriptors

    @property
    def global_features(self) -> Dict[str, TarHandler]:
        """
        :return: the global features tar handlers
        """
        return self._global_features

    @global_features.setter
    def global_features(self, global_features: Dict[str, TarHandler]):
        if not isinstance(global_features, dict):
            raise TypeError('dict of TarHandler expected')
        self._global_features = global_features

    @property
    def matches(self) -> Dict[str, TarHandler]:
        """
        :return: the matches tar handlers
        """
        return self._matches

    @matches.setter
    def matches(self, matches: Dict[str, TarHandler]):
        if not isinstance(matches, dict):
            raise TypeError('dict of TarHandler expected')
        self._matches = matches


KAPTURE_TARABLE_TYPES = {
    kapture.Keypoints,
    kapture.Descriptors,
    kapture.GlobalFeatures,
    kapture.Matches
}

FEATURES_TAR_FILENAMES = {
    kapture.Keypoints: lambda x: path.join('reconstruction', 'keypoints', x, 'keypoints.tar'),
    kapture.Descriptors: lambda x: path.join('reconstruction', 'descriptors', x, 'descriptors.tar'),
    kapture.GlobalFeatures: lambda x: path.join('reconstruction', 'global_features', x, 'global_features.tar'),
    kapture.Matches: lambda x: path.join('reconstruction', 'matches', x, 'matches.tar'),
}


def get_feature_tar_fullpath(kapture_type: Any, feature_name: str, kapture_dirpath: str = '') -> str:
    """
    Returns the full path to tar kapture file for a given datastructure and root directory.
    This path is the concatenation of the kapture root path and subpath into kapture into data structure.

    :param kapture_type: type of kapture data (kapture.RecordsCamera, kapture.Trajectories, ...)
    :param kapture_dirpath: root kapture path
    :return: full path of tar file for that type of data
    """
    assert kapture_type in FEATURES_TAR_FILENAMES
    filename = FEATURES_TAR_FILENAMES[kapture_type](feature_name)
    return path.join(kapture_dirpath, filename)


def list_files_in_tar(tar_handler: TarHandler,
                      filename_extensions: Optional[Union[str, List[str]]] = None) -> Iterable[str]:
    """
    Returns the list of file path into the given tar.
    If a list of extensions is given, returns only files with those extensions.

    :param tar_handler: opened tar reference
    :param filename_extensions: optional file name extensions to filter in
    :return: list of paths
    """
    # list all files
    file_paths = tar_handler.content.keys()

    # if extensions are given, keep only files that complies.
    if filename_extensions:
        if not isinstance(filename_extensions, list):
            # make sure extensions is a list
            filename_extensions = [filename_extensions]
        # make sure given extensions are lower case.
        filename_extensions = [ext.lower() for ext in filename_extensions]
        # check the extension is authorized
        file_paths = (
            file_path for file_path in file_paths
            if path.splitext(file_path)[1].lower() in filename_extensions
        )

    file_paths = (path_secure(file_path)
                  for file_path in file_paths)
    return file_paths


def retrieve_tar_handler_from_collection(kapture_type: Type[Union[kapture.Keypoints,
                                                                  kapture.Descriptors,
                                                                  kapture.GlobalFeatures,
                                                                  kapture.Matches]],
                                         feature_type: str,
                                         tar_handlers: Optional[Union[TarCollection,
                                                                      TarHandler]] = None) -> Optional[TarHandler]:
    """
    get a tar_handler from a collection of tar_handlers. the result will be None if it's not in a Tar

    :param kapture_type: kapture class type.
    :param feature_type: the name of the features type
    :param tar_handler: None or collection of preloaded tar archives, defaults to None
    :return: a TarHandler if the combo kapture_type/feature_type is in a tar, or None if it's in a directory
    """
    tar_local_handler = None
    if tar_handlers is not None:
        if isinstance(tar_handlers, TarCollection):
            if kapture_type == kapture.Keypoints and feature_type in tar_handlers.keypoints:
                tar_local_handler = tar_handlers.keypoints[feature_type]
            elif kapture_type == kapture.Descriptors and feature_type in tar_handlers.descriptors:
                tar_local_handler = tar_handlers.descriptors[feature_type]
            elif kapture_type == kapture.GlobalFeatures and feature_type in tar_handlers.global_features:
                tar_local_handler = tar_handlers.global_features[feature_type]
            elif kapture_type == kapture.Matches and feature_type in tar_handlers.matches:
                tar_local_handler = tar_handlers.matches[feature_type]
        elif isinstance(tar_handlers, TarHandler):
            tar_local_handler = tar_handlers
        else:
            raise TypeError(f'unknown {tar_handlers}')
    return tar_local_handler
